'''ODS class and save/load from pickle routines

-------
'''

from __future__ import print_function, division, unicode_literals

from .omas_utils import *

__version__ = open(os.path.abspath(str(os.path.dirname(__file__)) + os.sep + 'version'), 'r').read().strip()

__all__ = [
    'ODS', 'ODX',
    'CodeParameters', 'codeparams_xml_save', 'codeparams_xml_load',
    'ods_sample', 'different_ods',
    'save_omas_pkl', 'load_omas_pkl', 'through_omas_pkl',
    'save_omas_json', 'load_omas_json', 'through_omas_json',
    'save_omas_mongo', 'load_omas_mongo', 'through_omas_mongo',
    'save_omas_hdc', 'load_omas_hdc', 'through_omas_hdc',
    'load_omas_uda',
    'save_omas_nc', 'load_omas_nc', 'through_omas_nc',
    'save_omas_h5', 'load_omas_h5', 'through_omas_h5',
    'save_omas_ds', 'load_omas_ds', 'through_omas_ds',
    'load_omas_dx', 'save_omas_dx', 'through_omas_dx', 'ods_2_odx', 'odx_2_ods',
    'save_omas_imas', 'load_omas_imas', 'through_omas_imas', 'load_omas_iter_scenario', 'browse_imas',
    'save_omas_s3', 'load_omas_s3', 'through_omas_s3', 'list_omas_s3', 'del_omas_s3',
    'generate_xml_schemas', 'create_json_structure', 'create_html_documentation', 'symlink_imas_structure_versions',
    'imas_json_dir', 'imas_versions', 'IMAS_versions', 'omas_info', 'omas_info_node', 'get_actor_io_ids',
    'omas_rcparams', 'rcparams_environment', '__version__'
]

# List of functions that can be added by third-party Python
# packages for processing input data that will go in an ODS
# This is necessary because ODSs should only contain [int (arrays), floats (arrays), strings]
# It is used for example by OMFIT to process OMFITexpressions
input_data_process_functions = []

def force_imas_type(value):
    '''
    IMAS supports (arrays of) integers, floats and strings

    :param value: input value

    :return: input value converted to be IMAS compatible
    '''
    # lists are saved as numpy arrays, and 0D numpy arrays as scalars
    for function in input_data_process_functions:
        value = function(value)
    if isinstance(value, list):
        value = numpy.array(value)
    if isinstance(value, xarray.DataArray):
        value = value.values
    if isinstance(value, numpy.ndarray) and not len(value.shape):
        value = value.item()
    if isinstance(value, (numpy.string_, numpy.unicode_, numpy.str_)):
        value = value.item()
    elif isinstance(value, (float, numpy.floating)):
        value = float(value)
    elif isinstance(value, (int, numpy.integer)):
        value = int(value)
    if isinstance(value, basestring):
        pass
    elif isinstance(value, bytes):
        value = value.decode('utf-8', errors='ignore')
    return value


def consistency_checker(location, value, info, consistency_check):
    '''
    Print warnings or raise errors if object does not satisfy IMAS data dictionary
    Converts numeric data to INT/FLOAT depending on IMAS specifications

    :param value: value to check consistency of

    :param info: output of omas_info_node

    :param consistency_check: True, False, 'warn'

    :return: value
    '''
    # force type consistent with data dictionary
    if numpy.atleast_1d(is_uncertain(value)).any() or 'data_type' not in info:
        pass
    elif isinstance(value, numpy.ndarray):
        if 'FLT' in info['data_type']:
            value = value.astype(float)
        elif 'INT' in info['data_type']:
            value = value.astype(int)
    elif isinstance(value, (int, float, numpy.integer, numpy.floating)):
        if 'FLT' in info['data_type']:
            value = float(value)
        elif 'INT' in info['data_type']:
            value = int(value)
    # check type
    if not (isinstance(value, (int, float, unicode, str, numpy.ndarray, uncertainties.core.Variable)) or value is None or isinstance(value, CodeParameters)):
        text = 'Trying to write %s in %s\nSupported types are: string, float, int, array' % (type(value), location)
        if consistency_check == 'warn':
            printe(text)
        else:
            raise ValueError(text)
    # check consistency for scalar entries
    if 'data_type' in info and '_0D' in info['data_type'] and isinstance(value, numpy.ndarray):
        text = '%s must be a scalar of type %s' % (location, info['data_type'])
        if consistency_check == 'warn':
            printe(text)
        else:
            raise ValueError(text)
    # check consistency for number of dimensions
    elif 'coordinates' in info and len(info['coordinates']) and (not isinstance(value, numpy.ndarray) or len(value.shape) != len(info['coordinates'])):
        text = '%s must be an array with dimensions: %s' % (location, info['coordinates'])
        if consistency_check == 'warn':
            printe(text)
        else:
            raise ValueError(text)
    elif 'lifecycle_status' in info and info['lifecycle_status'] in ['obsolescent']:
        printe('%s is in %s state' % (location, info['lifecycle_status'].upper()))
    return value


class ODS(MutableMapping):
    """
    OMAS Data Structure class
    """

    def __init__(self,
                 imas_version=omas_rcparams['default_imas_version'],
                 consistency_check=omas_rcparams['consistency_check'],
                 dynamic_path_creation=omas_rcparams['dynamic_path_creation'],
                 location='',
                 cocos=omas_rcparams['cocos'],
                 cocosio=omas_rcparams['cocosio'],
                 coordsio=omas_rcparams['coordsio'],
                 unitsio=omas_rcparams['unitsio'],
                 structure=None):
        """
        :param imas_version: IMAS version to use as a constrain for the nodes names

        :param consistency_check: whether to enforce consistency with IMAS schema

        :param dynamic_path_creation: whether to dynamically create the path when setting an item
                                      * False: raise an error when trying to access a structure element that does not exists
                                      * True (default): arrays of structures can be incrementally extended by accessing at the next element in the array
                                      * 'dynamic_array_structures': arrays of structures can be dynamically extended

        :param location: string with location of this object relative to IMAS schema in ODS path format

        :param cocos: internal COCOS representation (this can only be set when the object is created)

        :param cocosio: COCOS representation of the data that is read/written from/to the ODS

        :param coordsio: ODS with coordinates to use for the data that is read/written from/to the ODS

        :param unitsio: ODS will return data with units if True

        :param structure: IMAS schema to use
        """
        if structure is None:
            structure = {}
        self.structure = structure
        self.omas_data = None
        self._consistency_check = consistency_check
        self._dynamic_path_creation = dynamic_path_creation
        if consistency_check and imas_version not in imas_versions:
            raise ValueError("Unrecognized IMAS version `%s`. Possible options are:\n%s" % (imas_version, imas_versions.keys()))
        self.imas_version = imas_version
        self.location = location
        self._cocos = cocos
        self.cocosio = cocosio
        self.coordsio = coordsio
        self.unitsio = unitsio

    def homogeneous_time(self, key='', default=True):
        '''
        Dynamically evaluate whether time is homogeneous or not
        NOTE: this method does not read ods['ids_properties.homogeneous_time'] instead it uses the time info to figure it out

        :param default: what to return in case no time basis is defined

        :return: True/False or default value (True) if no time basis is defined
        '''
        if not len(self.location) and not len(key):
            raise ValueError('homogeneous_time() can not be called on a top-level ODS')
        extra_info = {}
        self.time(key=key, extra_info=extra_info)
        homogeneous_time = extra_info['homogeneous_time']
        if homogeneous_time is None:
            return default
        else:
            return homogeneous_time

    def time(self, key='', extra_info=None, skip=None):
        """
        Return the time information for a given ODS location

        :param key: ods location

        :param extra_info: dictionary that will be filled in place with extra information about time

        :return: time information for a given ODS location (scalar or array)
        """

        if extra_info is None:
            extra_info = {}

        # subselect on requested key
        subtree = p2l(self.location + '.' + key)

        # get time nodes from data structure definitions
        loc = p2l(subtree)
        flat_structure, dict_structure = load_structure(loc[0], imas_version=self.imas_version)
        times_ds = [i2o(k) for k in flat_structure.keys() if k.endswith('.time')]

        # traverse ODS upstream until time information is found
        time = {}
        for sub in [subtree[:k] for k in range(len(subtree), 0, -1)]:
            times_sub_ds = [k for k in times_ds if k.startswith(l2u(sub))]
            subtree = l2o(sub)

            # get time data from ods
            times = {}
            n = len(self.location)
            for item in times_sub_ds:
                try:
                    time = self.__getitem__(u2o(item, subtree)[n:], None)  # traverse ODS
                    if isinstance(time, numpy.ndarray):
                        if time.size == 0:
                            continue
                        elif len(time.shape) > 1:
                            time = numpy.atleast_1d(numpy.squeeze(time))
                    times[item] = time
                except IndexError:
                    pass
                except ValueError as _excp:
                    if 'has no data' in repr(_excp):
                        pass
                    else:
                        # return False if time is not homogeneous
                        extra_info['homogeneous_time'] = False
                        return None
            times_values = list(times.values())

            extra_info['location'] = times.keys()
            # no time data defined
            if not len(times_values):
                time = None
                extra_info['homogeneous_time'] = None
            # if there is a single time entry, or there are multiple time entries that are all consistent with one another
            elif len(times) == 1 or all([times_values[0].shape == time.shape and numpy.allclose(times_values[0], time) for time in times_values[1:]]):
                time = times_values[0]
                extra_info['homogeneous_time'] = True
                return time
            # there are inconsistencies with different ways of specifying times in the IDS
            else:
                raise ValueError('Inconsistent time definitions in %s' % times.keys())

        return None

    def slice_at_time(self, time=None, time_index=None):
        '''
        method for selecting a time slice from an time-dependent ODS (NOTE: this method operates in place)

        :param time: time value to select

        :param time_index: time index to select (NOTE: time_index has precedence over time)

        :return: modified ODS
        '''

        # set time_index for parent and children
        if 'time' in self and isinstance(self['time'], numpy.ndarray):
            if time_index is None:
                time_index = numpy.argmin(abs(self['time'] - time))
                if (time - self['time'][time_index]) != 0.0:
                    printe('%s sliced at %s instead of requested time %s' % (self.location, self['time'][time_index], time))
                time = self['time'][time_index]
            if time is None:
                time = self['time'][time_index]

        # loop over items
        for item in self.keys():
            # time (if present) is treated last
            if item == 'time':
                continue

            # identify time-dependent data
            info = omas_info_node(o2u(self.ulocation + '.' + str(item)))
            if 'coordinates' in info and any([k.endswith('.time') for k in info['coordinates']]):

                # time-dependent arrays
                if not isinstance(self.getraw(item), ODS):
                    self[item] = numpy.atleast_1d(self[item][time_index])

                # time-depentend list of ODSs
                elif isinstance(self[item].omas_data, list) and len(self[item]) and 'time' in self[item][0]:
                    if time_index is None:
                        raise ValueError('`time` array is not set for `%s` ODS' % self.ulocation)
                    tmp = self[item][time_index]
                    self.getraw(item).clear()
                    self.getraw(item)[0] = tmp

            # go deeper inside ODSs that do not have time info
            elif isinstance(self.getraw(item), ODS):
                self.getraw(item).slice_at_time(time=time, time_index=time_index)

        # treat time
        if 'time' in self:
            self['time'] = numpy.atleast_1d(self['time'][time_index])

        return self

    @property
    def consistency_check(self):
        """
        property that sets whether consistency with IMAS schema is enabled or not

        :return: True/False
        """
        if not hasattr(self, '_consistency_check'):
            self._consistency_check = omas_rcparams['consistency_check']
        return self._consistency_check

    @consistency_check.setter
    def consistency_check(self, consistency_value):
        old_consistency_check = self._consistency_check
        try:
            # set ._consistency_check for this ODS
            self._consistency_check = consistency_value
            # set .consistency_check and assign the .structure and .location attributes to the underlying ODSs
            for item in self.keys():
                if isinstance(self.getraw(item), ODS) and 'code.parameters' in self.getraw(item).location:
                    continue
                else:
                    consistency_value_propagate = consistency_value
                    if consistency_value:
                        if isinstance(self.getraw(item), ODS) and not self.structure:
                            if not self.location:
                                # load the json structure file
                                structure = load_structure(item, imas_version=self.imas_version)[1][item]
                            else:
                                raise RuntimeError('When switching from False to True .consistency_check=True must be set at the top-level ODS')
                        else:
                            structure_key = item if not isinstance(item, int) else ':'
                            if structure_key in self.structure:
                                structure = self.structure[structure_key]
                            else:
                                options = list(self.structure.keys())
                                if len(options) == 1 and options[0] == ':':
                                    options = 'A numerical index is needed with n>=0'
                                else:
                                    options = 'Did you mean: %s' % options
                                spaces = ' ' * len('LookupError') + '  ' + ' ' * (len(self.location) + 2)
                                text = 'Not a valid IMAS %s location: %s' % (self.imas_version, self.location + '.' + structure_key)
                                if consistency_value == 'warn':
                                    printe(text)
                                    structure = {}
                                    consistency_value_propagate = False
                                else:
                                    raise LookupError(underline_last(text, len('LookupError: ')) + '\n' + options)
                        # assign structure and location information
                        if isinstance(self.getraw(item), ODS):
                            self.getraw(item).structure = structure
                            self.getraw(item).location = l2o([self.location] + [item])
                        else:
                            location = l2o([self.location] + [item])
                            info = omas_info_node(o2u(location), imas_version=self.imas_version)
                            value = consistency_checker(location, self.getraw(item), info, consistency_value)
                            if value is not self.getraw(item):
                                self.setraw(item, value)
                    # propagate consistency check
                    if isinstance(self.getraw(item), ODS):
                        self.getraw(item).consistency_check = consistency_value_propagate

        except Exception as _excp:
            # restore existing consistency_check value in case of error
            if old_consistency_check != consistency_value:
                for item in self.keys():
                    if isinstance(self.getraw(item), ODS):
                        self.getraw(item).consistency_check = old_consistency_check
                self.consistency_check = old_consistency_check
            raise  # (LookupError('Consistency check failed: %s' % repr(_excp)))

    @property
    def cocos(self):
        """
        property that tells in what COCOS format the data is stored internally of the ODS
        (NOTE: this parameter can only be set when the object is created)

        :return: cocos value
        """
        if not hasattr(self, '_cocos'):
            self._cocos = omas_rcparams['cocos']
        return self._cocos

    @cocos.setter
    def cocos(self, cocos_value):
        raise AttributeError('cocos parameter is readonly!')

    @property
    def cocosio(self):
        """
        property that tells in what COCOS format the data will be input/output

        :return: cocosio value
        """
        if not hasattr(self, '_cocosio'):
            self._cocosio = omas_rcparams['cocosio']
        return self._cocosio

    @cocosio.setter
    def cocosio(self, cocosio_value):
        self._cocosio = cocosio_value
        for item in self.keys():
            if isinstance(self.getraw(item), ODS):
                self.getraw(item).cocosio = cocosio_value

    @property
    def unitsio(self):
        """
        property that if data should be returned with units or not

        :return: unitsio value
        """
        if not hasattr(self, '_unitsio'):
            self._unitsio = omas_rcparams['unitsio']
        return self._unitsio

    @unitsio.setter
    def unitsio(self, unitsio_value):
        self._unitsio = unitsio_value
        for item in self.keys():
            if isinstance(self.getraw(item), ODS):
                self.getraw(item).unitsio = unitsio_value

    @property
    def coordsio(self):
        """
        property that tells in what COCOS format the data will be input/output

        :return: coordsio value
        """
        if not hasattr(self, '_coordsio'):
            self._coordsio = (None, omas_rcparams['coordsio'])
        return self._coordsio

    @coordsio.setter
    def coordsio(self, coordsio_value):
        if not isinstance(coordsio_value, (list, tuple)):
            coordsio_value = (self, coordsio_value)
        self._coordsio = coordsio_value
        for item in self.keys():
            if isinstance(self.getraw(item), ODS):
                self.getraw(item).coordsio = coordsio_value

    @property
    def dynamic_path_creation(self):
        """
        property that sets whether dynamic path creation is enabled or not

        :return: True/False
        """
        if not hasattr(self, '_dynamic_path_creation'):
            self._dynamic_path_creation = True
        return self._dynamic_path_creation

    @property
    def ulocation(self):
        '''
        :return: string with location of this object in universal ODS path format
        '''
        return o2u(self.location)

    @dynamic_path_creation.setter
    def dynamic_path_creation(self, dynamic_path_value):
        self._dynamic_path_creation = dynamic_path_value
        for item in self.keys():
            if isinstance(self.getraw(item), ODS):
                self.getraw(item).dynamic_path_creation = dynamic_path_value

    def _validate(self, value, structure):
        """
        Validate that the value is consistent with the provided structure field

        :param value: sub-tree to be checked

        :param structure: reference structure
        """
        for key in value.keys():
            structure_key = o2u(key)
            if isinstance(value[key], ODS) and value[key].consistency_check:
                value._validate(value[key], structure[structure_key])
            else:
                structure[structure_key]

    def set_child_locations(self):
        '''
        traverse ODSs and set .location attribute
        '''
        for item in self.keys():
            if isinstance(self.getraw(item), ODS):
                self.getraw(item).location = l2o([self.location, item])
                self.getraw(item).set_child_locations()

    def __setitem__(self, key, value):
        # handle individual keys as well as full paths
        key = p2l(key)

        if not len(key):
            return self

        # negative numbers are used to address arrays of structures from the end
        if isinstance(key[0], int) and key[0] < 0:
            if self.omas_data is None:
                key[0] = 0
            elif isinstance(self.omas_data, list):
                if not len(self.omas_data):
                    key[0] = 0
                else:
                    key[0] = len(self.omas_data) + key[0]
        # '+' is used to append new entry in array structure
        elif key[0] == '+':
            if self.omas_data is None:
                key[0] = 0
            elif isinstance(self.omas_data, list):
                key[0] = len(self.omas_data)

        # if the user has entered path rather than a single key
        if len(key) > 1:
            pass_on_value = value
            value = self.__class__(imas_version=self.imas_version,
                        consistency_check=self.consistency_check,
                        dynamic_path_creation=self.dynamic_path_creation,
                        cocos=self.cocos, cocosio=self.cocosio, coordsio=self.coordsio)

        # full path where we want to place the data
        location = l2o([self.location, key[0]])

        if self.consistency_check and '.code.parameters.' not in location:
            # perform consistency check with IMAS structure
            structure = {}
            structure_key = key[0] if not isinstance(key[0], int) else ':'
            try:
                if isinstance(value, ODS):
                    if not self.structure:
                        # load the json structure file
                        structure = load_structure(key[0], imas_version=self.imas_version)[1][key[0]]
                    else:
                        structure = self.structure[structure_key]
                        if not len(structure) and '.code.parameters' not in location:
                            raise ValueError('`%s` has no data' % location)
                    # check that tha data will go in the right place
                    self._validate(value, structure)
                    # assign structure and location information
                    value.structure = structure
                    # determine if entry is a list of structures or just a structure
                    if value.omas_data is None:
                        if ':' in value.structure.keys():
                            value.omas_data = []
                        elif len(value.structure.keys()):
                            value.omas_data = {}
                    value.location = location
                else:
                    self.structure[structure_key]

            except (LookupError, TypeError):
                text = 'Not a valid IMAS %s location: %s' % (self.imas_version, location)
                if self.consistency_check == 'warn':
                    printe(text)
                    if isinstance(value, ODS):
                        value.consistency_check = False
                elif self.consistency_check:
                    options = list(self.structure.keys())
                    if len(options) == 1 and options[0] == ':':
                        options = 'A numerical index is needed with n>=0'
                    else:
                        options = 'Did you mean: %s' % options
                    raise LookupError(underline_last(text, len('LookupError: ')) + '\n' + options)

        # check what container type is required and if necessary switch it
        if not self.omas_data or not len(self.omas_data):
            if isinstance(key[0], int):
                if not isinstance(self.omas_data, list):
                    self.omas_data = []
            else:
                if not isinstance(self.omas_data, dict):
                    self.omas_data = {}
        elif isinstance(key[0], int) and not isinstance(self.omas_data, list):
            raise TypeError('Cannot convert from dict to list once ODS has data')
        elif isinstance(key[0], basestring) and not isinstance(self.omas_data, dict):
            raise TypeError('Cannot convert from list to dict once ODS has data')

        # if the value is not an ODS strucutre
        if not isinstance(value, ODS):

            # convert simple dict of code.parameters to CodeParameters instances
            if '.code.parameters' in location and not isinstance(value, CodeParameters) and isinstance(value, (dict, ODS)):
                tmp = value
                value = CodeParameters()
                value.update(tmp)

            # now that all checks are completed we can assign the structure information
            if self.consistency_check and '.code.parameters.' not in location:
                ulocation = o2u(location)

                # handle cocos transformations coming in
                if self.cocosio and self.cocosio != self.cocos and '.' in location and ulocation in omas_physics.cocos_signals and not isinstance(value, ODS):
                    transform = omas_physics.cocos_signals[ulocation]
                    if transform == '?':
                        if self.consistency_check == 'warn':
                            printe('COCOS translation has not been setup: %s' % ulocation)
                            norm = 1.0
                        else:
                            raise ValueError('COCOS translation has not been setup: %s' % ulocation)
                    else:
                        norm = omas_physics.cocos_transform(self.cocosio, self.cocos)[transform]
                    value = value * norm

                # get node information
                info = omas_info_node(ulocation, imas_version=self.imas_version)

                # handle units (Python pint package)
                if str(value.__class__).startswith("<class 'pint."):
                    import pint
                    if 'units' in info and isinstance(value, pint.quantity._Quantity) or (isinstance(value, numpy.ndarray) and value.size and isinstance(numpy.atleast_1d(value).flat[0], pint.quantity._Quantity)):
                        value = value.to(info['units']).magnitude

                # coordinates interpolation
                ods_coordinates, input_coordinates = self.coordsio
                if input_coordinates:
                    all_coordinates = []
                    coordinates = []
                    if len(input_coordinates) and 'coordinates' in info:
                        all_coordinates = list(map(lambda x: u2o(x, self.location), info['coordinates']))
                        coordinates = list(filter(lambda coord: not coord.startswith('1...'), all_coordinates))
                    if len(coordinates):
                        # add any missing coordinate that were input
                        for coordinate in coordinates:
                            if coordinate not in ods_coordinates and coordinate in input_coordinates:
                                printd('Adding %s coordinate to ods' % (coordinate), topic='coordsio')
                                ods_coordinates[coordinate] = input_coordinates.__getitem__(coordinate, False)

                        # if all coordinates information is present
                        if all([coord in input_coordinates and coord in ods_coordinates for coord in coordinates]):
                            # if there is any coordinate that does not match
                            if any([len(input_coordinates.__getitem__(coord, None)) != len(ods_coordinates.__getitem__(coord, None)) or
                                    (not numpy.allclose(input_coordinates.__getitem__(coord, False), ods_coordinates.__getitem__(coord, False))) for coord in coordinates]):

                                # for the time being omas interpolates only 1D quantities
                                if len(info['coordinates']) > 1:
                                    raise Exception('coordio does not support multi-dimentional interpolation just yet')

                                # if the (first) coordinate is in input_coordinates
                                coordinate = coordinates[0]
                                if len(input_coordinates.__getitem__(coordinate, None)) != len(value):
                                    raise Exception('coordsio %s.shape=%s does not match %s.shape=%s' % (coordinate, input_coordinates.__getitem__(coordinate, False).shape, location, value.shape))
                                printd('Adding %s interpolated to input %s coordinate' % (self.location, coordinate), topic='coordsio')
                                value = omas_interp1d(ods_coordinates.__getitem__(coordinate, None), input_coordinates.__getitem__(coordinate, None), value)
                            else:
                                printd('%s ods and coordsio match' % (coordinates), topic='coordsio')
                        else:
                            printd('Adding `%s` without knowing coordinates `%s`' % (self.location, all_coordinates), topic='coordsio')

                    elif ulocation in omas_coordinates(self.imas_version) and location in ods_coordinates:
                        value = ods_coordinates.__getitem__(location, None)

            # lists are saved as numpy arrays, and 0D numpy arrays as scalars
            value = force_imas_type(value)

            # check that dimensions and data types are consistent with IMAS specifications
            if self.consistency_check and '.code.parameters.' not in location:
                value = consistency_checker(location, value, info, self.consistency_check)

        # check if the branch/node was dynamically created
        dynamically_created = False
        if key[0] not in self.keys() and len(key) > 1:
            dynamically_created = True

        # assign values to this ODS
        if key[0] not in self.keys() or len(key) == 1:
            # structure
            if isinstance(key[0], basestring):
                self.omas_data[key[0]] = value
            # arrays of structures
            else:
                # dynamic array structure creation
                if key[0] >= len(self.omas_data) and self.dynamic_path_creation == 'dynamic_array_structures':
                    for item in range(len(self.omas_data), key[0]):
                        ods = self.__class__()
                        ods.copy_attrs_from(self)
                        self[item] = ods
                # index exists
                if key[0] < len(self.omas_data):
                    self.omas_data[key[0]] = value
                # next index creation
                elif key[0] == len(self.omas_data) and self.dynamic_path_creation:
                    self.omas_data.append(value)
                # missing index
                else:
                    if not len(self.omas_data):
                        raise IndexError('`%s[%d]` but ods has no data' % (self.location, key[0]))
                    else:
                        raise IndexError('`%s[%d]` but maximun index is %d' % (self.location, key[0], len(self.omas_data) - 1))

        # pass the value one level deeper
        # and cleanup dynamically created branches if necessary (eg. if consistency check fails)
        if len(key) > 1:
            try:
                self.getraw(key[0])[key[1:]] = pass_on_value
            except LookupError:
                if dynamically_created:
                    del self[key[0]]
                raise

        # if the value is an ODS strucutre
        if isinstance(value, ODS) and value.omas_data is not None and len(value):
            # make sure entries have the right location
            self.set_child_locations()

    def getraw(self, key):
        '''
        Method to access data stored in ODS with no processing of the key, and it is thus faster than the ODS.__getitem__(key)
        Effectively behaves like a pure Python dictionary/list __getitem__.
        This method is mostly meant to be used in the inner workings of the ODS class.
        NOTE: ODS.__getitem__(key, False) can be used to access items in the ODS with disabled cocos and coordinates processing but with support for different syntaxes to access data

        :param key: string or integer

        :return: ODS value
        '''

        return self.omas_data[key]

    def setraw(self, key, value):
        '''
        Method to assign data to an ODS with no processing of the key, and it is thus faster than the ODS.__setitem__(key, value)
        Effectively behaves like a pure Python dictionary/list __setitem__.
        This method is mostly meant to be used in the inner workings of the ODS class.

        :param key: string or integer

        :param value: value to assign

        :return: value
        '''

        self.omas_data[key] = value
        return value

    def __getitem__(self, key, cocos_and_coords=True):
        '''
        ODS getitem method allows support for different syntaxes to access data

        :param key: different syntaxes to access data, for example:
              * ods['equilibrium']['time_slice'][0]['profiles_2d'][0]['psi']   # standard Python dictionary syntax
              * ods['equilibrium.time_slice[0].profiles_2d[0].psi']            # IMAS hierarchical tree syntax
              * ods['equilibrium.time_slice.0.profiles_2d.0.psi']              # dot separated string syntax
              * ods[['equilibrium','time_slice',0,'profiles_2d',0,'psi']]      # list of nodes syntax

        :param cocos_and_coords: processing of cocos transforms and coordinates interpolations [True/False/None]
              * True: enabled COCOS and enabled interpolation
              * False: enabled COCOS and disabled interpolation
              * None: disabled COCOS and disabled interpolation

        :return: ODS value
        '''

        # handle pattern match
        if isinstance(key, basestring) and key.startswith('@'):
            key = self.search_paths(key, 1, '@')[0]

        # handle individual keys as well as full paths
        key = p2l(key)

        if not len(key):
            return self

        # negative numbers are used to address arrays of structures from the end
        if isinstance(key[0], int) and key[0] < 0:
            if self.omas_data is None:
                key[0] = 0
            elif isinstance(self.omas_data, list):
                if not len(self.omas_data):
                    key[0] = 0
                else:
                    key[0] = len(self.omas_data) + key[0]

        dynamically_created = False

        # data slicing
        if key[0] == ':':
            data = []
            for k, item in enumerate(self.keys()):
                try:
                    data.append(self.__getitem__([item] + key[1:], cocos_and_coords))
                except ValueError:
                    data.append([])
            # handle missing data by filling out with NaNs
            valid = _empty = []
            for k, item in enumerate(data):
                if (isinstance(item, list) and not len(item)) or (isinstance(item, numpy.ndarray) and not item.size):
                    _empty.append(k)
                else:
                    valid = item
            if valid is not _empty and len(_empty):
                for k in _empty:
                    data[k] = valid * numpy.nan
            # force dtype to avoid obtaining arrays of objects in case
            # the shape of the concatenated arrays do not match
            return numpy.array(data, dtype=numpy.array(data[0]).dtype)

        # dynamic path creation
        elif key[0] not in self.keys():
            if self.dynamic_path_creation:
                dynamically_created = True
                self.__setitem__(key[0], self.__class__(imas_version=self.imas_version,
                                             consistency_check=self.consistency_check,
                                             dynamic_path_creation=self.dynamic_path_creation,
                                             cocos=self.cocos, cocosio=self.cocosio, coordsio=self.coordsio))
            else:
                location = l2o([self.location, key[0]])
                raise LookupError('Dynamic path creation is disabled, hence `%s` needs to be manually created' % location)

        value = self.omas_data[key[0]]
        if len(key) > 1:
            # if the user has entered path rather than a single key
            try:
                if isinstance(value, ODS):
                    return value.__getitem__(key[1:], cocos_and_coords)
                else:
                    return value[l2o(key[1:])]
            except ValueError:
                if dynamically_created:
                    del self[key[0]]
                raise
        else:

            if cocos_and_coords is not None and self.consistency_check and not isinstance(value, ODS):

                location = l2o([self.location, key[0]])
                ulocation = o2u(location)

                # handle cocos transformations going out
                if self.cocosio and self.cocosio != self.cocos and '.' in location and ulocation in omas_physics.cocos_signals:
                    transform = omas_physics.cocos_signals[ulocation]
                    if transform == '?':
                        if self.consistency_check == 'warn':
                            printe('COCOS translation has not been setup: %s' % ulocation)
                            norm = 1.0
                        else:
                            raise ValueError('COCOS translation has not been setup: %s' % ulocation)
                    else:
                        norm = omas_physics.cocos_transform(self.cocos, self.cocosio)[transform]
                    value = value * norm

                # get node information
                info = omas_info_node(ulocation, imas_version=self.imas_version)

                # coordinates interpolation
                ods_coordinates, output_coordinates = self.coordsio
                if cocos_and_coords and output_coordinates:
                    all_coordinates = []
                    coordinates = []
                    if len(output_coordinates) and 'coordinates' in info:
                        all_coordinates = list(map(lambda x: u2o(x, self.location), info['coordinates']))
                        coordinates = list(filter(lambda coord: not coord.startswith('1...'), all_coordinates))
                    if len(coordinates):
                        # if all coordinates information is present
                        if all([coord in output_coordinates and coord in ods_coordinates for coord in coordinates]):
                            # if there is any coordinate that does not match
                            if any([len(output_coordinates.__getitem__(coord, None)) != len(ods_coordinates.__getitem__(coord, None)) or
                                    (not numpy.allclose(output_coordinates.__getitem__(coord, None), ods_coordinates.__getitem__(coord, None))) for coord in coordinates]):

                                # for the time being omas interpolates only 1D quantities
                                if len(info['coordinates']) > 1:
                                    raise Exception('coordio does not support multi-dimentional interpolation just yet')

                                # if the (first) coordinate is in output_coordinates
                                coordinate = coordinates[0]
                                if len(ods_coordinates.__getitem__(coordinate, None)) != len(value):
                                    raise Exception('coordsio %s.shape=%s does not match %s.shape=%s' % (coordinate, output_coordinates.__getitem__(coordinate, False).shape, location, value.shape))
                                printd('Returning %s interpolated to output %s coordinate' % (location, coordinate), topic='coordsio')
                                try:
                                    value = omas_interp1d(output_coordinates.__getitem__(coordinate, None), ods_coordinates.__getitem__(coordinate, None), value)
                                except TypeError:
                                    if numpy.atleast_1d(is_uncertain(value)).any():
                                        v = omas_interp1d(output_coordinates.__getitem__(coordinate, None), ods_coordinates.__getitem__(coordinate, None), nominal_values(value))
                                        s = omas_interp1d(output_coordinates.__getitem__(coordinate, None), ods_coordinates.__getitem__(coordinate, None), std_devs(value))
                                        value = unumpy.uarray(v, s)
                            else:
                                printd('%s ods and coordsio match' % (coordinates), topic='coordsio')
                        else:
                            printd('Getting `%s` without knowing some of the coordinates `%s`' % (self.location, all_coordinates), topic='coordsio')

                    elif ulocation in omas_coordinates(self.imas_version) and location in output_coordinates:
                        value = output_coordinates.__getitem__(location, False)

                # handle units (Python pint package)
                if 'units' in info and self.unitsio:
                    import pint
                    from .omas_setup import ureg
                    if ureg[0] is None:
                        import pint
                        ureg[0] = pint.UnitRegistry()
                    value = value * getattr(ureg[0], info['units'])

            return value

    def __delitem__(self, key):
        # handle individual keys as well as full paths
        key = p2l(key)
        if len(key) > 1:
            # if the user has entered path rather than a single key
            del self.getraw(key[0])[key[1:]]
        else:
            return self.omas_data.__delitem__(key[0])

    def pretty_paths(self):
        """
        Traverse the ods and return paths that have data formatted nicely

        :return: list of paths that have data formatted nicely
        """
        return list(map(l2i, self.paths()))

    def paths(self, **kw):
        """
        Traverse the ods and return paths that have data

        :return: list of paths that have data
        """
        paths = kw.setdefault('paths', [])
        path = kw.setdefault('path', [])
        for kid in self.keys():
            if isinstance(self.getraw(kid), ODS):
                self.getraw(kid).paths(paths=paths, path=path + [kid])
            else:
                paths.append(path + [kid])
        return paths

    def full_paths(self):
        """
        Traverse the ods and return paths from root of ODS that have data

        :return: list of paths that have data
        """
        location = p2l(self.location)
        return [location + path for path in self.paths()]

    def flat(self):
        """
        :return: flat dictionary representation of the data
        """
        tmp = OrderedDict()
        for path in self.paths():
            tmp[l2o(path)] = self[path]
        return tmp

    def __getnewargs__(self):
        # tells pickle.dumps to pickle the omas object in such a way that a pickle.loads
        # back from that string will use omas.__new__ with consistency_check=False and dynamic_path_creation=True
        return (False, True)

    def __len__(self):
        return len(self.omas_data)

    def __iter__(self):
        return iter(self.keys())

    def __contains__(self, key):
        key = p2l(key)
        h = self
        for k in key:
            # h.omas_data is None when dict/list behaviour is not assigned
            if h.omas_data is not None and k in h.keys():
                h = h.__getitem__(k, False)
                continue  # continue to the next key
            else:
                return False
        # return False if checking existance of a leaf and the leaf exists but is unassigned
        if isinstance(h, ODS) and h.omas_data is None:
            return False
        return True

    def keys(self):
        if isinstance(self.omas_data, dict):
            # map keys with str to get strings and not unicode when working with Python 2.7
            return list(map(str, self.omas_data.keys()))
        elif isinstance(self.omas_data, list):
            return list(range(len(self.omas_data)))
        else:
            return []

    def values(self):
        return [self[item] for item in self.keys()]

    def __str__(self):
        return self.location

    def __repr__(self):
        return repr(self.omas_data)

    def __tree_repr__(self):
        '''
        OMFIT tree representation
        '''
        if not self.location:
            return self, []
        s = '--{%d}--' % len(self)
        if 'dataset_description.data_entry' in self:
            s += ' ' + ' '.join(['%s:%s' % (k, v) for k, v in self['dataset_description']['data_entry'].items() if v not in ['None', None]])
        if 'summary.ids_properties.comment' in self:
            s += ' ' + repr(self['summary.ids_properties.comment'])
        return s, []

    def get(self, key, default=None):
        """
        Check if key is present and if not return default value without creating value in omas data structure

        :param key: ods location

        :param default: default value

        :return: return default if key is not found
        """
        if key not in self:
            return default
        else:
            return self[key]

    def setdefault(self, key, value=None):
        """
        Set value if key is not present

        :param key: ods location

        :param value: value to set

        :return: value
        """
        if key not in self:
            self[key] = value
        return self[key]

    def __getstate__(self):
        state = {}
        for item in omas_dictstate:
            if item in self.__dict__:
                state[item] = self.__dict__[item]
        return state

    def copy(self):
        '''
        :return: copy.deepcopy of current ODS object
        '''
        return copy.deepcopy(self)

    def clear(self):
        '''
        remove data from a branch

        :return: current ODS object
        '''
        if isinstance(self.omas_data, dict):
            self.omas_data.clear()
        elif isinstance(self.omas_data, list):
            self.omas_data[:] = []
        return self

    def copy_attrs_from(self, ods):
        '''
        copy omas_ods_attrs ['_consistency_check','_dynamic_path_creation','imas_version','location','structure','_cocos','_cocosio','_coordsio'] attributes from input ods

        :param ods: input ods

        :return: self
        '''
        for item in omas_ods_attrs:
            setattr(self, item, getattr(ods, item))
        return self

    def prune(self):
        '''
        Prune ODS branches that are leafless

        :return: number of branches that were pruned
        '''
        n = 0
        for item in self.keys():
            if isinstance(self.getraw(item), ODS):
                n += self.getraw(item).prune()
                if not len(self.getraw(item).keys()):
                    n += 1
                    del self[item]
        return n

    def set_time_array(self, key, time_index, value):
        '''
        Convenience function for setting time dependent arrays

        :param key: ODS location to edit

        :param time_index: time index of the value to set

        :param value: value to set

        :return: time dependent array
        '''

        orig_value = []
        if key in self:
            orig_value = numpy.atleast_1d(self[key]).tolist()

        # substitute
        if time_index < len(orig_value):
            orig_value[time_index] = value
        # append
        elif time_index == len(orig_value):
            orig_value = orig_value + [value]
        else:
            key = p2l(key)
            raise IndexError('%s has length %d and time_index %d is bejond current range' % (l2o(key), len(orig_value), time_index))

        self[key] = numpy.atleast_1d(orig_value)
        return orig_value

    def update(self, ods2):
        '''
        Adds ods2's key-values pairs to the ods

        :param ods2: dictionary or ODS to be added into the ODS
        '''
        if isinstance(ods2, ODS):
            for item in ods2.paths():
                self[item] = ods2[item]
        else:
            try:
                bkp_dynamic_path_creation = self.dynamic_path_creation
                self.dynamic_path_creation = 'dynamic_array_structures'
                for item in ods2.keys():
                    self[item] = ods2[item]
            finally:
                self.dynamic_path_creation = bkp_dynamic_path_creation

    def list_coordinates(self):
        '''
        return dictionary with coordinates in a given ODS

        :return: dictionary with coordinates (keys are absolute location, values are relative locations)
        '''
        coords = {}

        n = len(self.location)
        for full_path in self.full_paths():
            if l2u(full_path) in omas_coordinates(self.imas_version):
                coords[l2o(full_path)] = self[l2o(full_path)[n:]]

        return coords

    def coordinates(self, key):
        '''
        return dictionary with coordinates of a given ODS location

        :param key: ODS location to return the coordinates of
                    Note: both the key location and coordinates must have data

        :return: OrderedDict with coordinates of a given ODS location
        '''
        coords = OrderedDict()

        self.__getitem__(key, False)  # raise an error if data is not there
        ulocation = l2u(p2l(key))
        info = omas_info_node(ulocation)
        if 'coordinates' not in info:
            raise ValueError('ODS location `%s` has no coordinates information' % ulocation)
        coordinates = map(lambda x: u2o(x, key), info['coordinates'])
        for coord in coordinates:
            coords[coord] = self[coord]  # this will raise an error if the coordinates data is not there

        return coords

    def search_paths(self, search_pattern, n=None, regular_expression_startswith=''):
        '''
        Find ODS locations that match a pattern

        :param search_pattern: regular expression ODS location string

        :param n: raise an error if a numbe of occurrences different from n is found

        :param regular_expression_startswith: indicates that use of regular expressions
               in the search_pattern is preceeded by certain characters.
               This is used internally by some methods of the ODS to force users
               to use '@' to indicate access to a path by regular expression.

        :return: list of ODS locations matching search_pattern pattern
        '''
        if not isinstance(search_pattern, basestring):
            return [search_pattern]

        elif regular_expression_startswith:
            if not search_pattern.startswith(regular_expression_startswith):
                return [search_pattern]
            else:
                search_pattern = search_pattern[len(regular_expression_startswith):]

        search = re.compile(search_pattern)
        matches = []
        for path in map(l2o, self.full_paths()):
            if re.match(search, path):
                matches.append(path)
        if n is not None and len(matches) != n:
            raise ValueError('Found %d matches of `%s` instead of the %d requested\n%s' % (len(matches), search_pattern, n, '\n'.join(matches)))
        return matches

    def xarray(self, key):
        '''
        Returns data of an ODS location and correspondnig coordinates as an xarray dataset
        Note that the Dataset and the DataArrays have their attributes set with the ODSs structure info

        :param key: ODS location

        :return: xarray dataset
        '''
        key = self.search_paths(key, 1, '@')[0]

        import xarray
        key = p2l(key)

        info = omas_info_node(l2u(key))
        coords = self.coordinates(key)

        short_coords = OrderedDict()
        for coord in coords:
            short_coords[p2l(coord)[-1]] = coords[coord]

        ds = xarray.Dataset()
        ds[key[-1]] = xarray.DataArray(self[key], coords=short_coords, dims=short_coords.keys(), attrs=info)
        ds.attrs['y'] = key[-1]
        ds.attrs['y_full'] = l2o(key)

        ds.attrs['x'] = []
        ds.attrs['x_full'] = []
        for coord in coords:
            info = omas_info_node(o2u(coord))
            ds[p2l(coord)[-1]] = xarray.DataArray(coords[coord], dims=p2l(coord)[-1], attrs=info)
            ds.attrs['x'].append(p2l(coord)[-1])
            ds.attrs['x_full'].append(coord)
        return ds

    def dataset(self, homogeneous=[False, 'time', 'full', None][-1]):
        '''
        Return xarray.Dataset representation of a whole ODS

        Forming the N-D labeled arrays (tensors) that are at the base of xarrays,
        requires that the number of elements in the arrays do not change across
        the arrays of data structures.

        :param homogeneous: * False: flat representation of the ODS
                                      (data is not collected across arrays of structures)
                            * 'time': collect arrays of structures only along the time dimension
                                      (always valid for homogeneous_time=True)
                            * 'full': collect arrays of structures along all dimensions
                                      (may be valid in many situations, especially related to
                                       simulation data with homogeneous_time=True and where
                                       for example number of ions, sources, etc. do not vary)
                            * None: smart setting, uses homogeneous='time' if homogeneous_time=True else False

        :return: xarray.Dataset
        '''
        import xarray

        if not self.location:
            DS = xarray.Dataset()
            for ds in self:
                DS.update(self[ds].dataset(homogeneous=homogeneous))
            return DS

        def arraystruct_indexnames(key):
            '''
            return list of strings with a name for each of the arrays of structures indexes

            :param key: ods location

            :return: list of strings
            '''
            base = key.split('.')[0]
            coordinates = []
            counter = 0
            for c in [':'.join(key.split(':')[:k + 1]).strip('.') for k, struct in enumerate(key.split(':'))]:
                info = omas_info_node(o2u(c))
                if 'coordinates' in info:
                    for infoc in info['coordinates']:
                        if infoc == '1...N':
                            infoc = c
                        coordinates.append('__'+'_'.join([base, infoc.split('.')[-1]] + [str(k) for k in p2l(key) if isinstance(k, int) or k == ':'] + ['index%d__' % counter]))
                        counter += 1
            return coordinates

        # Generate paths with ':' for the arrays of structures
        # that we want to collect across
        paths = self.paths()
        if self.location:
            fpaths = list(map(lambda key: [self.location] + key, paths))

        if homogeneous is None:
            homogeneous = 'time' if self.homogeneous_time() else False
        if not homogeneous:
            fupaths = list(map(l2o, fpaths))
        elif homogeneous == 'time':
            fupaths = numpy.unique(list(map(l2ut, fpaths)))
        elif homogeneous == 'full':
            fupaths = numpy.unique(list(map(l2u, fpaths)))
        else:
            raise ValueError("OMAS dataset homogeneous attribute can only be [False, None, 'time', 'full']")
        upaths = fupaths
        if self.location:
            n = len(self.location)
            upaths = list(map(lambda key: key[n + 1:], fupaths))

        # Figure out coordinate indexes
        # NOTE: We use coordinates indexes instead of proper coordinates
        #       since in IMAS these are time dependent quantities
        #       Eg. 'equilibrium.time_slice[:].profiles_1d.psi'
        coordinates = {}
        for fukey, ukey in zip(fupaths, upaths):
            coordinates[fukey] = arraystruct_indexnames(fukey)

        # Generate dataset
        DS = xarray.Dataset()
        for fukey, ukey in zip(fupaths, upaths):
            if not len(omas_info_node(o2u(fukey))):
                printe('WARNING: %s is not part of IMAS' % o2i(fukey))
                continue
            data = self[ukey]  # OMAS data slicing at work
            for k, c in enumerate(coordinates[fukey]):
                if c not in DS:
                    DS[c] = xarray.DataArray(numpy.arange(data.shape[k]), dims=c)
            try:
                DS[fukey] = xarray.DataArray(data, dims=coordinates[fukey])
            except Exception:
                printe('Error with %s with coordinates %s' % (fukey, coordinates[fukey]))
                raise
        return DS

    def satisfy_imas_requirements(self, attempt_fix=True, raise_errors=True):
        """
        Assign .time and .ids_properties.homogeneous_time info for top-level structures
        since these are required for writing an IDS to IMAS

        :param attempt_fix: fix dataset_description and wall IDS to have 0 times if none is set

        :param raise_errors: raise errors if could not satisfy IMAS requirements

        :return: `True` if all is good, `False` if requirements are not satisfied, `None` if fixes were applied
        """

        # if called at top level, loop over all data structures
        if not len(self.location):
            out = []
            for ds in self:
                out.append(self.getraw(ds).satisfy_imas_requirements(attempt_fix=attempt_fix, raise_errors=raise_errors))
            if any([k is False for k in out]):
                return False
            elif any([k is None for k in out]):
                return None
            else:
                return True

        ds = p2l(self.location)[0]

        if ds in add_datastructures:
            return True

        extra_info = {}
        time = self.time(extra_info=extra_info)
        if extra_info['homogeneous_time'] is False:
            self['ids_properties']['homogeneous_time'] = extra_info['homogeneous_time']
        elif time is not None and len(time):
            self['time'] = time
            self['ids_properties']['homogeneous_time'] = extra_info['homogeneous_time']
        elif attempt_fix and ds in ['dataset_description', 'wall']:
            self['time'] = [0.0]
            extra_info['homogeneous_time'] = True
            self['ids_properties']['homogeneous_time'] = extra_info['homogeneous_time']
            return None
        elif raise_errors:
            raise ValueError(self.location + '.time cannot be automatically filled! Missing time information in the data structure.')
        else:
            return False
        return True

    def save(self, *args, **kw):
        r"""
        Save OMAS data

        :param filename: filename.XXX where the extension is used to select save format method (eg. 'pkl','nc','h5','ds')
                         set to `imas`, `s3`, `hdc`, `mongo` for load methods that do not have a filename with extension

        :param \*args: extra arguments passed to save_omas_XXX() method

        :param \**kw: extra keywords passed to save_omas_XXX() method

        :return: return from save_omas_XXX() method
        """
        if '.' not in args[0]:
            ext = args[0]
            args = args[1:]
        else:
            ext = os.path.splitext(args[0])[-1].strip('.')
        return eval('save_omas_' + ext)(self, *args, **kw)

    def load(self, *args, **kw):
        r"""
        Load OMAS data

        :param filename: filename.XXX where the extension is used to select load format method (eg. 'pkl','nc','h5','ds')
                         set to `imas`, `s3`, `hdc`, `mongo` for save methods that do not have a filename with extension

        :param \*args: extra arguments passed to load_omas_XXX() method

        :param \**kw: extra keywords passed to load_omas_XXX() method

        :return: ODS with loaded data
        """
        if '.' not in args[0]:
            ext = args[0]
            args = args[1:]
        else:
            ext = os.path.splitext(args[0])[-1].strip('.')
        if 'consistency_check' in kw:
            consistency_check = kw['consistency_check']
        else:
            consistency_check = self.consistency_check
        if self.location:
            kw['consistency_check'] = False
        else:
            kw['consistency_check'] = consistency_check
        results = eval('load_omas_' + ext)(*args, **kw)
        if ext in ['mongo']:
            if not len(results):
                raise RuntimeError(ext + ' query returned no result!')
            elif len(results) > 1:
                raise RuntimeError(ext + ' query returned more than one result!')
            else:
                self.omas_data = list(results.values())[0].omas_data
        else:
            self.omas_data = results.omas_data
        self.consistency_check = consistency_check
        return self

    def diff(self, ods, ignore_type=False, ignore_empty=False):
        '''
        return differences between this ODS and the one passed

        :param ods: ODS to compare against

        :param ignore_type: ignore object type differences

        :param ignore_empty: ignore empty nodes

        :return: dictionary with differences
        '''
        return different_ods(self, ods, ignore_type=ignore_type, ignore_empty=ignore_empty)

    def from_structure(self, structure):
        '''
        Generate an ODS starting from a hierarchical structure made of dictionaries and lists

        :param structure: input structure

        :return: ODS
        '''
        if isinstance(structure, dict):
            keys = list(map(str, structure.keys()))
        elif isinstance(structure, list):
            keys = list(range(len(structure)))
        else:
            raise ValueError('from_structure must be fed either a structure made of dictionaries and lists')

        for item in keys:
            if isinstance(structure[item], dict):
                self[item].from_structure(structure[item])
            elif isinstance(structure[item], list):
                # identify if this is a leaf
                if len(structure[item]) and not isinstance(structure[item][0], dict):
                    self[item] = numpy.array(structure[item])
                # or a node in the IMAS tree
                else:
                    self[item].from_structure(structure[item])
            else:
                self[item] = copy.deepcopy(structure[item])

        return self

    def codeparams2xml(self):
        '''
        Convert code.parameters to a XML string
        '''
        if not self.location:
            for item in self:
                self[item].codeparams2xml()
            return
        elif ('code.parameters' in self and isinstance(self['code.parameters'], CodeParameters)):
            self['code.parameters'] = self['code.parameters'].tostring()
        elif ('parameters' in self and isinstance(self['parameters'], CodeParameters)):
            self['parameters'] = self['parameters'].tostring()

    def codeparams2dict(self):
        '''
        Convert code.parameters to a CodeParameters dictionary object
        '''
        if not self.location:
            for item in self:
                self[item].codeparams2dict()
            return
        try:
            if ('code.parameters' in self and isinstance(self['code.parameters'], basestring)):
                self['code.parameters'] = CodeParameters().fromstring(self['code.parameters'])
            elif ('parameters' in self and isinstance(self['parameters'], basestring)):
                self['parameters'] = CodeParameters().fromstring(self['parameters'])
        except Exception as _excp:
            printe(repr(_excp))


class CodeParameters(dict):
    """
    Class used to interface with IMAS code-parameters XML files
    """
    def __init__(self, string=None):
        if isinstance(string, basestring):
            if os.path.exists(string):
                self.fromfile(string)
            else:
                self.fromstring(string)

    def fromstring(self, code_params_string):
        '''
        Load data from code.parameters XML string

        :param code_params_string: XML string

        :return: self
        '''
        import xmltodict
        self.clear()
        tmp = xmltodict.parse(code_params_string)['parameters']
        recursive_interpreter(tmp)
        self.update(tmp)
        return self

    def fromfile(self, code_params_file):
        '''
        Load data from code.parameters XML file

        :param code_params_file: XML file

        :return: self
        '''
        with open(code_params_file, 'r') as f:
            return self.fromstring(f.read())

    def tostring(self):
        '''
        generate an XML string from this dictionary

        :return: XML string
        '''
        import xmltodict
        tmp = {'parameters': OrderedDict()}
        tmp['parameters'].update(copy.deepcopy(self))
        recursive_encoder(tmp)
        return xmltodict.unparse(tmp, pretty=True)


def codeparams_xml_save(f):
    '''
    Decorator function to be used around the omas_save_XXX methods to enable
    saving of code.parameters as an XML string
    '''
    def wrapper(ods, *args, **kwargs):
        with omas_environment(ods, xmlcodeparams=True):
            return f(ods, *args, **kwargs)
    return wrapper


def codeparams_xml_load(f):
    '''
    Decorator function to be used around the omas_load_XXX methods to enable
    loading of code.parameters from an XML string
    '''
    def wrapper(*args, **kwargs):
        ods = f(*args, **kwargs)
        ods.codeparams2dict()
        return ods
    return wrapper

# --------------------------------------------
# import sample functions and add them as ODS methods
# --------------------------------------------
try:
    from . import omas_sample
    from .omas_sample import ods_sample

    __all__.append('omas_sample')
    for item in omas_sample.__ods__:
        setattr(ODS, 'sample_' + item, getattr(omas_sample, item))
except ImportError as _excp:
    printe('OMAS sample function are not available: ' + repr(_excp))
    raise

# --------------------------------------------
# import physics functions and add them as ODS methods
# --------------------------------------------
try:
    from . import omas_physics
    from .omas_physics import *

    __all__.append('omas_physics')
    __all__.extend(omas_physics.__all__)
    for item in omas_physics.__ods__:
        setattr(ODS, 'physics_' + item, getattr(omas_physics, item))
except ImportError as _excp:
    printe('OMAS physics function are not available: ' + repr(_excp))

# --------------------------------------------
# import plotting functions and add them as ODS methods
try:
    # --------------------------------------------
    from . import omas_plot

    __all__.append('omas_plot')
    for item in omas_plot.__ods__:
        setattr(ODS, 'plot_' + item, getattr(omas_plot, item))
except ImportError as _excp:
    printe('OMAS plotting function are not available: ' + repr(_excp))

omas_ods_attrs = ['_consistency_check', '_dynamic_path_creation', 'imas_version', 'location', 'structure', '_cocos', '_cocosio', '_coordsio']
omas_dictstate = dir(ODS)
omas_dictstate.extend(['omas_data'] + omas_ods_attrs)
omas_dictstate = sorted(list(set(omas_dictstate)))


# --------------------------------------------
# save and load OMAS with Python pickle
# --------------------------------------------
def save_omas_pkl(ods, filename, **kw):
    """
    Save OMAS data set to Python pickle

    :param ods: OMAS data set

    :param filename: filename to save to

    :param kw: keywords passed to pickle.dump function
    """
    printd('Saving to %s' % filename, topic='pkl')

    kw.setdefault('protocol', omas_rcparams['pickle_protocol'])

    with open(filename, 'wb') as f:
        pickle.dump(ods, f, **kw)


def load_omas_pkl(filename, consistency_check=None, imas_version=None):
    """
    Load OMAS data set from Python pickle

    :param filename: filename to save to

    :param consistency_check: verify that data is consistent with IMAS schema (skip if None)

    :param imas_version: imas version to use for consistency check (leave original if None)

    :returns: ods OMAS data set
    """
    printd('Loading from %s' % filename, topic='pkl')

    with open(filename, 'rb') as f:
        tmp = pickle.load(f)
    if imas_version is not None:
        tmp.imas_version = imas_version
    if consistency_check is not None:
        tmp.consistency_check = consistency_check
    return tmp


def through_omas_pkl(ods):
    """
    Test save and load Python pickle

    :param ods: ods

    :return: ods
    """
    if not os.path.exists(tempfile.gettempdir() + '/OMAS_TESTS/'):
        os.makedirs(tempfile.gettempdir() + '/OMAS_TESTS/')
    filename = tempfile.gettempdir() + '/OMAS_TESTS/test.pkl'
    save_omas_pkl(ods, filename)
    ods1 = load_omas_pkl(filename)
    return ods1


# --------------------------------------------
# import other omas tools and methods in this namespace
# --------------------------------------------
from .omas_imas import *
from .omas_s3 import *
from .omas_nc import *
from .omas_json import *
from .omas_structure import *
from .omas_hdc import *
from .omas_uda import *
from .omas_h5 import *
from .omas_ds import *
from .omas_mongo import *


# --------------------------------------------
# backward compatibility
# --------------------------------------------
def save_omas_itm(*args, **kw):
    '''
    Dummy function to avoid older versions of OMFIT not to start with latest OMAS
    '''
    raise NotImplementedError('omas itm functions are deprecated')


load_omas_itm = through_omas_itm = save_omas_itm
__all__.extend(['save_omas_itm', 'load_omas_itm', 'through_omas_itm'])



'''save/load from NC routines

-------
'''

from __future__ import print_function, division, unicode_literals

from .omas_utils import *
from .omas_core import ODS


# --------------------------------------------
# save and load OMAS with NetCDF
# --------------------------------------------
def save_omas_nc(ods, filename, **kw):
    """
    Save an OMAS data set to NetCDF file

    :param ods: OMAS data set

    :param filename: filename to save to

    :param kw: arguments passed to the netCDF4 Dataset function
    """
    printd('Saving to %s' % filename, topic='nc')

    from netCDF4 import Dataset
    odsf = ods.flat()
    with Dataset(filename, 'w', **kw) as dataset:
        for item in odsf:
            dims = []
            data = numpy.asarray(odsf[item])
            std = None
            tmp = is_uncertain(odsf[item])
            if numpy.any(numpy.atleast_1d(tmp)):
                std = std_devs(data)
                data = nominal_values(data)
            for k in range(len(numpy.asarray(odsf[item]).shape)):
                dims.append('dim_%d' % (data.shape[k]))
                if dims[-1] not in dataset.dimensions:
                    dataset.createDimension(dims[-1], data.shape[k])
            if std is None:
                dataset.createVariable(item, data.dtype, dims)[:] = data
            else:
                dataset.createVariable(item, data.dtype, dims)[:] = data
                dataset.createVariable(item + '_error_upper', data.dtype, dims)[:] = std


def load_omas_nc(filename, consistency_check=True, ):
    """
    Load an OMAS data set from NetCDF file

    :param filename: filename to load from

    :param consistency_check: verify that data is consistent with IMAS schema

    :return: OMAS data set
    """
    printd('Loading from %s' % filename, topic='nc')

    from netCDF4 import Dataset
    ods = ODS(consistency_check=False)
    with Dataset(filename, 'r') as dataset:
        for item in dataset.variables.keys():
            if item.endswith('_error_upper'):
                continue
            if dataset.variables[item].shape:
                # arrays
                if item + '_error_upper' in dataset.variables.keys():
                    ods[item] = uarray(numpy.array(dataset.variables[item]),
                                       numpy.array(dataset.variables[item + '_error_upper']))
                else:
                    ods[item] = numpy.array(dataset.variables[item])
            else:
                # uncertain scalars
                if item + '_error_upper' in dataset.variables.keys():
                    ods[item] = ufloat(dataset.variables[item][0].item(),
                                       dataset.variables[item + '_error_upper'][0].item())
                else:
                    try:
                        # scalars
                        ods[item] = dataset.variables[item][0].item()
                    except AttributeError:
                        # strings
                        ods[item] = dataset.variables[item][0]
    ods.consistency_check = consistency_check
    return ods


def through_omas_nc(ods, method=['function', 'class_method'][1]):
    """
    Test save and load NetCDF

    :param ods: ods

    :return: ods
    """
    if not os.path.exists(tempfile.gettempdir() + '/OMAS_TESTS/'):
        os.makedirs(tempfile.gettempdir() + '/OMAS_TESTS/')
    filename = tempfile.gettempdir() + '/OMAS_TESTS/test.nc'

    if method == 'function':
        save_omas_json(ods, filename)
        ods1 = load_omas_json(filename)
    else:
        ods.save(filename)
        ods1 = ODS().load(filename)
    return ods1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Coordinates interpolation
=========================
This example illustrates how OMAS can automatically interpolate data
defined on coordinates that were already present in the data structure.
This feature is extremely useful when different codes that have different
computational grids need to read/write parts of the same data structure.
"""

from __future__ import print_function, division, unicode_literals

import numpy
from omas import *

print('A')
# SET: if a coordinate exists, then that is the coordinate that it is used
ods1 = ODS()
ods1['equilibrium.time_slice[0].profiles_1d.psi'] = numpy.linspace(0, 1, 10)
with omas_environment(ods1, coordsio={'equilibrium.time_slice[0].profiles_1d.psi': numpy.linspace(0, 1, 5)}):
    ods1['equilibrium.time_slice[0].profiles_1d.f'] = numpy.linspace(0, 1, 5)
print('ods1 f has length %d' % len(ods1['equilibrium.time_slice[0].profiles_1d.f']))
print()

###################################
# A prints the following::
#
#     A
#     ods1 f has length 10
#

print('B')
# SET: if a does not exists, then that coordinate is set
ods2 = ODS()
with omas_environment(ods2, coordsio={'equilibrium.time_slice[0].profiles_1d.psi': numpy.linspace(0, 1, 5)}):
    ods2['equilibrium.time_slice[0].profiles_1d.pressure'] = numpy.linspace(0, 1, 5)
print('ods2 p has length %d' % len(ods2['equilibrium.time_slice[0].profiles_1d.pressure']))
print()

###################################
# B the following::
#
#     B
#     ods2 p has length 5
#

print('C')
# SET: coordinates can be taken from existing ODSs
ods3 = ODS()
with omas_environment(ods3, coordsio=ods1):
    ods3['equilibrium.time_slice[0].profiles_1d.f'] = ods1['equilibrium.time_slice[0].profiles_1d.f']
with omas_environment(ods3, coordsio=ods2):
    ods3['equilibrium.time_slice[0].profiles_1d.pressure'] = ods2['equilibrium.time_slice[0].profiles_1d.pressure']
print('ods3 f has length %d' % len(ods3['equilibrium.time_slice[0].profiles_1d.f']))
print('ods3 p has length %d' % len(ods3['equilibrium.time_slice[0].profiles_1d.pressure']))
print()

###################################
# C prints the following::
#
#     C
#     ods3 f has length 10
#     ods3 p has length 10
#

print('D')
# SET: order matters
ods4 = ODS()
with omas_environment(ods4, coordsio=ods2):
    ods4['equilibrium.time_slice[0].profiles_1d.pressure'] = ods2['equilibrium.time_slice[0].profiles_1d.pressure']
with omas_environment(ods4, coordsio=ods1):
    ods4['equilibrium.time_slice[0].profiles_1d.f'] = ods1['equilibrium.time_slice[0].profiles_1d.f']
print('ods4 f has length %d' % len(ods4['equilibrium.time_slice[0].profiles_1d.f']))
print('ods4 p has length %d' % len(ods4['equilibrium.time_slice[0].profiles_1d.pressure']))
print()

###################################
# D prints the following::
#
#     D
#     ods4 f has length 5
#     ods4 p has length 5
#

print('E')
# GET: ods can be queried on different coordinates
with omas_environment(ods4, coordsio=ods1):
    print('ods4 f is returned with length %d' % len(ods4['equilibrium.time_slice[0].profiles_1d.f']))
print('ods4 f is returned with length %d' % len(ods4['equilibrium.time_slice[0].profiles_1d.f']))

###################################
# E prints the following::
#
#     E
#     ods4 f is returned with length 10
#     ods4 f is returned with length 5
#



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time in OMAS
============
This example illustrates how OMAS can handle the time dimension
"""

from __future__ import print_function, division, unicode_literals

from omas import *

# test generation of a sample ods
ods = ODS()
ods['equilibrium.time_slice'][0]['time'] = 100
ods['equilibrium.time_slice.0.global_quantities.ip'] = 0.0
ods['equilibrium.time_slice'][1]['time'] = 200
ods['equilibrium.time_slice.1.global_quantities.ip'] = 1.0
ods['equilibrium.time_slice'][2]['time'] = 300
ods['equilibrium.time_slice.1.global_quantities.ip'] = 2.0

# get time information from children
extra_info = {}
print(ods.time('equilibrium', extra_info=extra_info))
# >>  [100 200 300]
print(extra_info)
# >>  {'location': 'equilibrium.time_slice.:.time', 'homogeneous_time': True}

# show use of .homogeneous_time property
print(ods.homogeneous_time('equilibrium'))
# >>  True

# time arrays can be set using `set_time_array` function
# this simplifies the logic in the code since one does not
# have to check if the array was already there or not
ods.set_time_array('equilibrium.time', 0, 101)
ods.set_time_array('equilibrium.time', 1, 201)
ods.set_time_array('equilibrium.time', 2, 302)

# the make the timeslices consistent
ods['equilibrium.time_slice'][0]['time'] = 101
ods['equilibrium.time_slice'][1]['time'] = 201
ods['equilibrium.time_slice'][2]['time'] = 302

# get time information from explicitly set time array
extra_info = {}
print(ods.time('equilibrium', extra_info=extra_info))
# >>  [101 201 301]
print(extra_info)
# >>  {'location': '.time', 'homogeneous_time': True}

# get time value from item in array of structures
extra_info = {}
print(ods['equilibrium.time_slice'][0].time(extra_info=extra_info))
# >>  101
print(extra_info)
# >>  {'location': 'equilibrium.time_slice.0.time', 'homogeneous_time': True}

# get time array from array of structures
extra_info = {}
print(ods['equilibrium.time_slice'].time(extra_info=extra_info))
# >>  [101 201 302]
print(extra_info)
# >>  {'location': 'equilibrium.time_slice.:.time', 'homogeneous_time': True}

# get time from parent
extra_info = {}
print(ods.time('equilibrium.time_slice.0.global_quantities.ip', extra_info=extra_info))
# >>  101
print(extra_info)
# >>  {'location': '.time', 'homogeneous_time': None}

# slice whole ODS at given time
ods1 = ods['equilibrium'].slice_at_time(101)
print(ods.time('equilibrium'))
# >>  [101]



import os
import sys
import glob
import subprocess

all_install_require = ['numpy', 'uncertainties', 'pint', 'netCDF4', 'boto3', 'matplotlib', 'scipy', 'h5py', 'pymongo', 'dnspython','xmltodict']
install_requires = {}
# https://github.com/pydata/xarray/commit/faacc8da000b7971233142be349ee39c6d088510
install_requires[2] = all_install_require + ['xarray<=0.11.0']
install_requires[3] = all_install_require + ['xarray']

extras_require = {'hdc': ['pyhdc'],
                  'imas': ['imas'],
                  'uda': ['pyuda'],
                  'build_structures': ['bs4'],
                  'build_documentation': ['Sphinx', 'sphinx-bootstrap-theme', 'sphinx-gallery', 'Pillow']}

# Add .json IMAS structure files to the package
here = os.path.abspath(os.path.split(__file__)[0]) + os.sep

# Automatically generate requirement.txt file if this is the OMAS repo and requirements.txt is missing
if os.path.exists(here + '.git') and not os.path.exists(here + 'requirements.txt'):
    for version in [2, 3]:
        with open(here + 'requirements_python%d.txt' % version, 'w') as f:
            f.write('# Do not edit this file by hand, operate on setup.py instead\n#\n')
            f.write('# usage: pip install -r requirements_python%d.txt\n\n' % version)
            for item in install_requires[version]:
                f.write(item.ljust(25) + '# required\n')
            f.write('\n')
            for requirement in extras_require:
                for item in extras_require[requirement]:
                    if requirement in ['imas', 'hdc', 'uda', 'build_structures']:
                        item = '#' + item
                    f.write(item.ljust(25) + '# %s\n' % requirement)

packages = ['omas',
            'omas.examples',
            'omas.samples',
            'omas.tests',
            'omas.utilities']
package_data = {'omas': ['*.py', 'version'],
                'omas.examples': ['*.py'],
                'omas.samples': ['*'],
                'omas.tests': ['*.py'],
                'omas.utilities': ['*.py']}
for item in glob.glob(os.sep.join([here, 'omas', 'imas_structures', '*'])):
    packages.append('omas.imas_structures.' + os.path.split(item)[1])
    package_data['omas.imas_structures.' + os.path.split(item)[1]] = ['*.json']

long_description = '''
OMAS is a Python library designed to simplify the interface of third-party codes with the `ITER <http://iter.org>`_ Integrated Modeling and Analysis Suite (`IMAS <https://confluence.iter.org/display/IMP>`_).

* It provides a **convenient Python API**

* capable of storing data with **different file/database formats**

* in a form that is **always compatible with the IMAS data model**

Mapping the physics codes I/O to the IMAS data model is done in third party Python codes such as the `OMFIT framework <http://gafusion.github.io/OMFIT-source>`_.
'''

print('INFO: run the `imports_check.py` script to quickly verify that all Python dependencies for OMAS are installed\n')

from setuptools import setup

setup(
    name='omas',
    version=open(here + 'omas/version', 'r').read().strip(),
    description='Ordered Multidimensional Array Structures',
    url='https://gafusion.github.io/omas',
    author='Orso Meneghini',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ],
    keywords='integrated modeling OMFIT IMAS ITER',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=packages,
    package_data=package_data,
    install_requires=install_requires[sys.version_info[0]],
    extras_require=extras_require)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Rapid plot
==========
This example shows how to use the .plot_quantity() method to
conveniently and rapidly plot individual ODS quantities
"""

from __future__ import print_function, division, unicode_literals

from matplotlib import pyplot
import os
from omas import *

ods = ods_sample()

kw = {}
kw['lw'] = 2

fig = pyplot.figure(figsize=(12, 6))

ax = fig.add_subplot(2, 2, 1)
t = '$T_e$'
ods.plot_quantity('@core.*elec.*tem', t, **kw)
ax.set_title(t, y=0.8)

ax = fig.add_subplot(2, 2, 3, sharex=ax)
t = '$T_i$'
ods.plot_quantity('@core.*ion.0.*tem', t, **kw)
ax.set_title(t, y=0.8)

ax = fig.add_subplot(4, 2, 2, sharex=ax)
t = '$n_e$'
ods.plot_quantity('@core.*elec.*dens', t, **kw)
ax.set_title(t, y=0.8)

ax = fig.add_subplot(4, 2, 4, sharex=ax)
t = '$n_D$'
ods.plot_quantity('@core.*ion.0.*dens.*th', t, **kw)
ax.set_title(t, y=0.8)

ax = fig.add_subplot(4, 2, 6, sharex=ax)
t = '$n_C$'
ods.plot_quantity('@core.*ion.1.*dens.*th', t, **kw)
ax.set_title(t, y=0.8)

ax = fig.add_subplot(4, 2, 8, sharex=ax)
t = '$Z_{eff}$'
ods.plot_quantity('@core.*zeff', t, **kw)
ax.set_title(t, y=0.8)

fig.subplots_adjust(hspace=0)
pyplot.show()



#!/usr/bin/env python
# # -*- coding: utf-8 -*-

"""
Test script for omas/omas_physics.py

.. code-block:: none

   python -m unittest omas/tests/test_omas_physics

-------
"""

# Basic imports
from __future__ import print_function, division, unicode_literals
import os
import unittest
import numpy
import warnings
import copy
import itertools

# Plot imports
import matplotlib as mpl
from matplotlib import pyplot as plt

# OMAS imports
from omas import *
from omas.omas_utils import *
from omas.omas_physics import *

try:
    import pint

    failed_PINT = False
except ImportError as _excp:
    failed_PINT = _excp


class TestOmasPhysics(unittest.TestCase):
    """
    Test suite for omas_physics.py
    """

    # Flags to edit while testing
    verbose = False  # Spammy, but occasionally useful for debugging a weird problem

    # Utilities for this test
    def printv(self, *arg):
        """Utility for tests to use"""
        if self.verbose:
            print(*arg)

    def setUp(self):
        test_id = self.id()
        test_name = '.'.join(test_id.split('.')[-2:])
        self.printv('{}...'.format(test_name))

    def tearDown(self):
        test_name = '.'.join(self.id().split('.')[-2:])
        self.printv('    {} done.'.format(test_name))

    def test_core_profiles_pressures(self):
        ods = ODS()
        ods.sample_core_profiles(include_pressure=False)
        ods2 = core_profiles_pressures(ods, update=True)
        diff = ods.diff(ods2)
        assert (not diff)

        ods2 = core_profiles_pressures(ods, update=False)
        assert (all(['press' in item for item in ods2.flat().keys() if not item.endswith('rho_tor_norm')]))

    def test_core_profiles_currents(self):

        rho = numpy.linspace(0., 1., 4)
        Jval = 1e5 * numpy.ones(4)
        jdef = {}
        Js = ['j_actuator', 'j_bootstrap', 'j_non_inductive', 'j_ohmic', 'j_total']
        for j in Js:
            jdef[j] = 'default'

        def CPC(ods, kw=jdef, should_RE=False, should_AE=False, warn=False):

            try:
                core_profiles_currents(ods, 0, rho, warn=warn, **kw)
            except RuntimeError as err:
                if should_RE:
                    pass
                else:
                    print(repr(kw))
                    raise err
            except AssertionError as err:
                if should_AE:
                    pass
                else:
                    print(repr(kw))
                    raise err
            else:
                if should_RE:
                    raise RuntimeError("Should have raised RuntimeError but didn't: " + repr(kw))
                elif should_AE:
                    raise AssertionError("Should have raised AssertionError but didn't: " + repr(kw))
            return

        # Try just setting one
        for i, J1 in enumerate(Js):
            kw = copy.deepcopy(jdef)
            kw[J1] = Jval
            CPC(ODS(), kw=kw, should_RE=(J1 == 'j_actuator'))

            # Now try setting two
            for J2 in Js[i + 1:]:
                kw = copy.deepcopy(jdef)
                kw[J1] = Jval
                kw[J2] = Jval
                should_RE = ((not isinstance(kw['j_actuator'], basestring) or kw['j_actuator'] != 'default') and
                             (isinstance(kw['j_bootstrap'], basestring) and kw['j_bootstrap'] == 'default') and
                             (isinstance(kw['j_non_inductive'], basestring) and kw['j_non_inductive'] == 'default'))
                CPC(ODS(), kw=kw, should_RE=should_RE)

        # Try setting three
        for keys in list(itertools.combinations(Js, 3)):
            kw = copy.deepcopy(jdef)
            for key in keys:
                kw[key] = Jval
            if (('j_actuator' in keys) and
                    ('j_bootstrap' in keys) and
                    ('j_non_inductive' in keys)):
                CPC(ODS(), kw=kw, should_AE=True)
                kw['j_non_inductive'] = 2 * Jval
            elif (('j_non_inductive' in keys) and
                  ('j_ohmic' in keys) and
                  ('j_total' in keys)):
                CPC(ODS(), kw=kw, should_AE=True)
                kw['j_total'] = 2 * Jval
            CPC(ODS(), kw=kw)

        # Try setting four
        for dkey, rkey, factor in [('j_total', 'j_non_inductive', 2),
                                   ('j_ohmic', 'j_non_inductive', 2),
                                   ('j_non_inductive', 'j_total', 3),
                                   ('j_bootstrap', 'j_total', 2),
                                   ('j_actuator', 'j_total', 2)]:
            kw = copy.deepcopy(jdef)
            for key in kw.keys():
                if key != dkey:
                    kw[key] = Jval
            CPC(ODS(), kw=kw, should_AE=True)
            kw[rkey] = factor * Jval
            CPC(ODS(), kw=kw)

        # Try all 5
        kw = copy.deepcopy(jdef)
        for key in Js:
            kw[key] = Jval
        CPC(ODS(), kw=kw, should_AE=True)
        kw['j_non_inductive'] = 2 * Jval
        CPC(ODS(), kw=kw, should_AE=True)
        kw['j_total'] = 3 * Jval
        CPC(ODS(), kw=kw, warn=True)  # just to cover the warn sections

        # Now test with equilibrium and existing quantities
        ods = ODS().sample_equilibrium()
        kw = {'j_actuator': Jval,
              'j_bootstrap': Jval}
        CPC(ods, kw=kw)  # j_ni = 2
        kw = {'j_bootstrap': 2 * Jval}
        CPC(ods, kw=kw, should_AE=True)
        kw = {'j_bootstrap': 2 * Jval, 'j_non_inductive': None}
        CPC(ods, kw=kw)  # j_ni = 3
        kw = {'j_actuator': 1.5 * Jval,
              'j_bootstrap': 1.5 * Jval}
        CPC(ods, kw=kw)
        kw = {'j_bootstrap': 2 * Jval, 'j_actuator': None}
        CPC(ods, kw=kw)

        kw = {'j_ohmic': Jval}
        CPC(ods, kw=kw)  # j_total is 4
        kw = {'j_ohmic': 2 * Jval}
        CPC(ods, kw=kw, should_AE=True)
        kw = {'j_ohmic': 2 * Jval, 'j_total': None}
        CPC(ods, kw=kw)  # j_total is 5
        kw = {'j_ohmic': Jval, 'j_non_inductive': None}
        CPC(ods, kw=kw, should_AE=True)
        kw = {'j_ohmic': Jval, 'j_non_inductive': None, 'j_actuator': None}
        CPC(ods, kw=kw)  # j_ni is 4

    def test_current_from_eq(self):
        ods = ODS().sample_equilibrium()
        current_from_eq(ods, 0)

    def test_define_cocos(self):
        cocos_none = define_cocos(None)
        cocos1 = define_cocos(1)
        cocos2 = define_cocos(2)
        cocos3 = define_cocos(3)
        cocos4 = define_cocos(4)
        cocos5 = define_cocos(5)
        cocos6 = define_cocos(6)
        cocos7 = define_cocos(7)
        cocos8 = define_cocos(8)
        cocos11 = define_cocos(11)
        for cocos in [cocos_none, cocos1, cocos2, cocos5, cocos6, cocos11]:
            assert cocos['sigma_Bp'] == 1
        for cocos in [cocos3, cocos4, cocos7, cocos8]:
            assert cocos['sigma_Bp'] == -1

    def test_cocos_transform(self):
        assert cocos_transform(None, None)['TOR'] == 1
        assert cocos_transform(1, 3)['POL'] == -1
        for cocos_ind in range(1, 9):
            assert cocos_transform(cocos_ind, cocos_ind + 10)['invPSI'] != 1
            for cocos_add in range(2):
                for thing in ['BT', 'TOR', 'POL', 'Q']:
                    assert cocos_transform(cocos_ind + cocos_add * 10, cocos_ind + cocos_add * 10)[thing] == 1

    def test_coordsio(self):
        data5 = numpy.linspace(0, 1, 5)
        data10 = numpy.linspace(0, 1, 10)

        if self.verbose:
            os.environ['OMAS_DEBUG_TOPIC'] = 'coordsio'

        # data can be entered without any coordinate checks
        ods0 = ODS()
        ods0['equilibrium.time_slice[0].profiles_1d.psi'] = data10
        ods0['equilibrium.time_slice[0].profiles_1d.f'] = data5
        assert (len(ods0['equilibrium.time_slice[0].profiles_1d.psi']) != len(ods0['equilibrium.time_slice[0].profiles_1d.f']))

        # if a coordinate exists, then that is the coordinate that it is used
        ods1 = ODS()
        ods1['equilibrium.time_slice[0].profiles_1d.psi'] = data10
        with omas_environment(ods1, coordsio={'equilibrium.time_slice[0].profiles_1d.psi': data5}):
            ods1['equilibrium.time_slice[0].profiles_1d.f'] = data5
        assert (len(ods1['equilibrium.time_slice[0].profiles_1d.f']) == 10)

        # if a coordinate does not exists, then it is added
        ods2 = ODS()
        with omas_environment(ods2, coordsio={'equilibrium.time_slice[0].profiles_1d.psi': data5}):
            ods2['equilibrium.time_slice[0].profiles_1d.pressure'] = data5
        assert (len(ods2['equilibrium.time_slice[0].profiles_1d.pressure']) == 5)

        # coordinates can be easily copied over from existing ODSs with .list_coordinates() method
        ods3 = ODS()
        ods3.update(ods1.list_coordinates())
        with omas_environment(ods3, coordsio=ods1):
            ods3['equilibrium.time_slice[0].profiles_1d.f'] = ods1['equilibrium.time_slice[0].profiles_1d.f']
        with omas_environment(ods3, coordsio=ods2):
            ods3['equilibrium.time_slice[0].profiles_1d.pressure'] = ods2['equilibrium.time_slice[0].profiles_1d.pressure']
        assert (len(ods3['equilibrium.time_slice[0].profiles_1d.f']) == 10)
        assert (len(ods3['equilibrium.time_slice[0].profiles_1d.pressure']) == 10)

        # ods can be queried on different coordinates than they were originally filled in (ods example)
        with omas_environment(ods3, coordsio=ods2):
            assert (len(ods3['equilibrium.time_slice[0].profiles_1d.f']) == 5)
        assert (len(ods3['equilibrium.time_slice[0].profiles_1d.f']) == 10)

        # ods can be queried on different coordinates than they were originally filled in (ods example)
        with omas_environment(ods3, coordsio=ods3):
            assert (len(ods3['equilibrium.time_slice[0].profiles_1d.f']) == 10)

        # ods can be queried on different coordinates than they were originally filled in (dictionary example)
        with omas_environment(ods3, coordsio={'equilibrium.time_slice[0].profiles_1d.psi': data5}):
            assert (len(ods3['equilibrium.time_slice[0].profiles_1d.f']) == 5)
        assert (len(ods3['equilibrium.time_slice[0].profiles_1d.f']) == 10)

        # this case is different because the coordinate and the data do not share the same parent
        ods5 = ODS()
        ods5['core_profiles.profiles_1d[0].grid.rho_tor_norm'] = data5
        with omas_environment(ods5, coordsio={'core_profiles.profiles_1d[0].grid.rho_tor_norm': data10}):
            ods5['core_profiles.profiles_1d[0].electrons.density_thermal'] = data10
        assert (len(ods5['core_profiles.profiles_1d[0].grid.rho_tor_norm']) == 5)
        assert (len(ods5['core_profiles.profiles_1d[0].electrons.density_thermal']) == 5)

        ods6 = ODS()
        ods6['core_profiles.profiles_1d[0].grid.rho_tor_norm'] = data5

    def test_cocosio(self):
        x = numpy.linspace(.1, 1, 10)

        ods = ODS(cocosio=11, cocos=11)
        ods['equilibrium.time_slice.0.profiles_1d.psi'] = x
        assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], x))

        ods = ODS(cocosio=11, cocos=2)
        ods['equilibrium.time_slice.0.profiles_1d.psi'] = x
        assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], x))

        ods = ODS(cocosio=2, cocos=11)
        ods['equilibrium.time_slice.0.profiles_1d.psi'] = x
        assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], x))

        ods = ODS(cocosio=2, cocos=2)
        ods['equilibrium.time_slice.0.profiles_1d.psi'] = x
        assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], x))

        # reassign the same value
        ods = ODS(cocosio=2)
        ods['equilibrium.time_slice.0.profiles_1d.psi'] = x
        ods['equilibrium.time_slice.0.profiles_1d.psi'] = ods['equilibrium.time_slice.0.profiles_1d.psi']
        assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], x))

        # use omas_environment
        ods = ODS(cocosio=2)
        ods['equilibrium.time_slice.0.profiles_1d.psi'] = x
        with omas_environment(ods, cocosio=11):
            assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], -x * (2 * numpy.pi)))

        ods['equilibrium.time_slice.0.profiles_1d.psi'] = x
        assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], x))

    def test_coordsio_cocosio(self):
        x = numpy.linspace(0.1, 1, 11)
        y = numpy.linspace(-1, 1, 11)

        xh = numpy.linspace(0.1, 1, 21)
        yh = numpy.linspace(-1, 1, 21)

        ods = ODS()
        with omas_environment(ods, cocosio=2):
            ods['equilibrium.time_slice.0.profiles_1d.psi'] = x
            assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], x))
        assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], -x * 2 * numpy.pi))

        with omas_environment(ods, cocosio=2, coordsio={'equilibrium.time_slice.0.profiles_1d.psi': xh}):
            ods['equilibrium.time_slice.0.profiles_1d.phi'] = yh
            assert (len(ods['equilibrium.time_slice.0.profiles_1d.phi']) == len(yh))
            assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], xh))
        assert (numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], -x * 2 * numpy.pi))
        assert (len(ods['equilibrium.time_slice.0.profiles_1d.phi']) == len(y))

        ods = ODS()

        p = 1 - numpy.linspace(0, 1, 10)
        psi2 = numpy.linspace(0, 1, 10) ** 2
        psi11 = -psi2 * (2 * numpy.pi)

        psi2_ = numpy.linspace(0, 1, 10)
        psi11_ = -psi2_ * (2 * numpy.pi)

        with omas_environment(ods, cocosio=2):
            ods['equilibrium.time_slice.0.profiles_1d.psi'] = psi2
            assert all(ods['equilibrium.time_slice.0.profiles_1d.psi'] == psi2)
        assert all(ods['equilibrium.time_slice.0.profiles_1d.psi'] == psi11)

        with omas_environment(ods, cocosio=2):
            assert all(ods['equilibrium.time_slice.0.profiles_1d.psi'] == psi2)
        assert all(ods['equilibrium.time_slice.0.profiles_1d.psi'] == psi11)

        with omas_environment(ods, cocosio=2, coordsio={'equilibrium.time_slice.0.profiles_1d.psi': psi2}):
            assert all(ods['equilibrium.time_slice.0.profiles_1d.psi'] == psi2), '2'
        assert all(ods['equilibrium.time_slice.0.profiles_1d.psi'] == psi11), '11'

        with omas_environment(ods, cocosio=2, coordsio={'equilibrium.time_slice.0.profiles_1d.psi': psi2}):
            ods['equilibrium.time_slice.0.profiles_1d.pressure'] = p
            assert all(ods['equilibrium.time_slice.0.profiles_1d.pressure'] == p), 'p2'
        assert all(ods['equilibrium.time_slice.0.profiles_1d.pressure'] == p), 'p11'

        psi2_ = numpy.linspace(0, 1, 10)
        psi11_ = -psi2_ * (2 * numpy.pi)

        with omas_environment(ods, cocosio=2, coordsio={'equilibrium.time_slice.0.profiles_1d.psi': psi2_}):
            assert numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.psi'], psi2_)
            assert numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.pressure'], numpy.interp(psi2_, psi2, p))
            index = numpy.argsort(psi11_)
            assert numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.pressure'], numpy.interp(psi11_, psi11[index], p[index]))

        psi2__ = numpy.linspace(0, 1, 100)
        psi11__ = -psi2__ * (2 * numpy.pi)

        with omas_environment(ods, cocosio=2, coordsio={'equilibrium.time_slice.0.profiles_1d.psi': psi2__}):
            assert numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.pressure'], numpy.interp(psi2__, psi2, p))
            index = numpy.argsort(psi11)
            assert numpy.allclose(ods['equilibrium.time_slice.0.profiles_1d.pressure'], numpy.interp(psi11__, psi11[index], p[index]))

    @unittest.skipUnless(not failed_PINT, str(failed_PINT))
    def test_handle_units(self):
        import pint
        ureg = pint.UnitRegistry()

        ods = ODS()
        ods['equilibrium.time_slice[0].constraints.diamagnetic_flux.time_measurement'] = 8.0 * ureg.milliseconds
        assert (ods['equilibrium.time_slice[0].constraints.diamagnetic_flux.time_measurement'] == 0.008)

        with omas_environment(ods, unitsio=True):
            tmp = ods['equilibrium.time_slice[0].constraints.diamagnetic_flux.time_measurement']
            assert (tmp.magnitude == 0.008)
            assert (tmp.units == 'second')

    def test_search_ion(self):
        ods = ODS()
        ods.sample_core_profiles(include_pressure=False)

        tmp = search_ion(ods["core_profiles.profiles_1d.0.ion"], 'D')
        assert repr(tmp) == '{0: [0]}'

        tmp = search_ion(ods["core_profiles.profiles_1d.0.ion"], multiple_matches_raise_error=False)
        assert repr(tmp) == '{0: [0], 1: [0]}'

        try:
            tmp = search_ion(ods["core_profiles.profiles_1d.0.ion"])
            raise AssertError('multiple_matches_raise_error')
        except IndexError:
            pass

        tmp = search_ion(ods["core_profiles.profiles_1d.0.ion"], A=12)
        assert repr(tmp) == '{1: [0]}'

        tmp = search_ion(ods["core_profiles.profiles_1d.0.ion"], 'W', no_matches_raise_error=False)
        assert repr(tmp) == '{}'

        try:
            tmp = search_ion(ods["core_profiles.profiles_1d.0.ion"], 'W')
            raise AssertError('no_matches_raise_error failed')
        except IndexError:
            pass

    def test_search_in_array_structure(self):
        ods = ODS()

        tmp = search_in_array_structure(ods['core_transport.model'], {'identifier.name': 'omas_tgyro'}, no_matches_return=0)[0]
        assert tmp == 0

        ods['core_transport.model.+.identifier.name'] = 'omas_tgyro'
        ods['core_transport.model.+.identifier.name'] = 'test1'
        ods['core_transport.model.+.identifier.name'] = 'omas_tgyro'
        ods['core_transport.model.-1.identifier.description'] = 'bla bla'
        ods['core_transport.model.+.identifier.name'] = 'test2'

        try:
            search_in_array_structure(ods['core_transport.model'], {'identifier.name': 'omas_tgyro'})
            raise AssertionError('multiple_matches_raise_error failed')
        except IndexError:
            pass

        tmp = search_in_array_structure(ods['core_transport.model'], {'identifier.name': 'omas_tgyro'}, multiple_matches_raise_error=False)
        assert tmp[0] == 0 and tmp[1] == 2

        tmp = search_in_array_structure(ods['core_transport.model'], {'identifier.name': 'omas_tgyro', 'identifier.description': 'bla bla'})
        assert tmp[0] == 2

    def test_latest_cocos(self):
        from omas.omas_utils import list_structures, omas_rcparams
        from omas.omas_physics import generate_cocos_signals

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=".*defining its COCOS transform.*")
            generate_cocos_signals(list_structures(imas_version=omas_rcparams['default_imas_version']), threshold=0, write=False, verbose=False)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestOmasPhysics)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
IMAS consistency check
======================
OMAS can enforce consistency with IMAS data structure.
There are three possible settings:

* **consistency_check = False**: no consistency check. `False` can be used to use OMAS for other purposes other than IMAS

* **consistency_check = 'warn'**: will print a wanrning if entry is outside of IMAS scope. `'warn'` can be useful when wanting to work with entries that are not yet supported by IMAS. Remember to open a JIRA issue (https://jira.iter.org) to start a conversation about making these entries officially part of IMAS.

* **consistency_check = True**: will raise an error if entry is outside of IMAS scope. Recommended when working with IMAS. The error raised will clearly say where the error is, and provide suggestions based on valid IMAS structure.
"""

from __future__ import print_function, division, unicode_literals

import os

from omas import *

print('*' * 20)
print('consistency_check = False')
print('*' * 20)
ods = ods_sample()
ods.consistency_check = False
# add entry with wrong dimensions
ods['equilibrium.time_slice[0].profiles_2d[0].b_tor'] = 1
# add entry that is obsolescent
ods['equilibrium.time_slice[0].profiles_2d[0].b_tor'] = [[1, 1], [1, 1]]
# add entry that does not exist in IMAS
ods['equilibrium.time_slice[2].does_not_exist.global_quantities.ip'] = 1

print('*' * 20)
print("consistency_check = 'warn'")
print('*' * 20)
ods = ods_sample()
ods.consistency_check = 'warn'
# add entry with wrong dimensions
ods['equilibrium.time_slice[0].profiles_2d[0].b_tor'] = 1
# add entry that is obsolescent
ods['equilibrium.time_slice[0].profiles_2d[0].b_tor'] = [[1, 1], [1, 1]]
# add entry that does not exist in IMAS
ods['equilibrium.time_slice[2].does_not_exist.global_quantities.ip'] = 1

###################################
# In the above example `consistency_check = warn` or `True` result in the following warnings::
#
#     equilibrium.time_slice.:.profiles_2d.:.b_tor must be an array with dimensions: ['equilibrium.time_slice[:].profiles_2d[:].grid.dim1', 'equilibrium.time_slice[:].profiles_2d[:].grid.dim2']
#     equilibrium.time_slice.:.profiles_2d.:.b_tor is in OBSOLESCENT state

print('*' * 20)
print("consistency_check = True")
print('*' * 20)
ods = ods_sample()
ods.consistency_check = True  # this is the default
# add entry with wrong dimensions
ods['equilibrium.time_slice[0].profiles_2d[0].b_tor'] = 1
# add entry that is obsolescent
ods['equilibrium.time_slice[0].profiles_2d[0].b_tor'] = [[1, 1], [1, 1]]
# add entry that does not exist in IMAS
ods['equilibrium.time_slice[2].does_not_exist.global_quantities.ip'] = 1

###################################
# In the above example `consistency_check = True` result in the following error::
#
#     LookupError: Not a valid IMAS 3.18.0 location: `equilibrium.time_slice.:.does_not_exist`
#                                                                              ^^^^^^^^^^^^^^
#     Did you mean: ['profiles_2d', 'ggd', 'boundary', 'profiles_1d', 'constraints', 'global_quantities', 'coordinate_system', 'boundary_separatrix', 'time', 'convergence']



'''functions for handling IMAS XML data dictionaries

-------
'''

from __future__ import print_function, division, unicode_literals

from .omas_utils import *


# --------------------------------------------
# generation of the imas structure json files
# IDS's XML documentation can be found where IMAS is installed at: $IMAS_PREFIX/share/doc/imas/html_documentation.html
# --------------------------------------------
def generate_xml_schemas(imas_version=None):
    """
    Generate IMAS IDSDef.xml files by:
     1. clone the IMAS data-dictionary repository (access to git.iter.org required)
     2. download the Saxon XSLT and XQuery Processor
     3. generate IDSDef.xml in imas_structures folder
    """
    import subprocess

    saxon_version = 'SaxonHE9-9-1-4J'
    saxon_major_version = '9.' + saxon_version.split('-')[1]

    # clone the IMAS data-dictionary repository
    dd_folder = os.sep.join([imas_json_dir, '..', 'data-dictionary'])
    if not os.path.exists(dd_folder):
        subprocess.Popen(
            'cd %s ; git clone ssh://git@git.iter.org/imas/data-dictionary.git' % os.sep.join([imas_json_dir, '..']),
            stdout=subprocess.PIPE, shell=True).communicate()[0]

    # download Saxon
    sax_folder = os.sep.join([imas_json_dir, '..', saxon_version])
    if not os.path.exists(sax_folder):
        subprocess.Popen("""
        cd {install_dir}
        curl https://iweb.dl.sourceforge.net/project/saxon/Saxon-HE/9.9/{saxon_version}.zip > {saxon_version}.zip
        unzip -d {saxon_version} {saxon_version}.zip
        rm {saxon_version}.zip""".format(install_dir=os.sep.join([imas_json_dir, '..']), saxon_major_version=saxon_major_version, saxon_version=saxon_version), shell=True).communicate()

    # fetch data structure updates
    subprocess.Popen("""
    cd {dd_folder}
    git fetch
    """.format(dd_folder=dd_folder), shell=True).communicate()

    # find IMAS data-dictionary tags
    result = b2s(subprocess.Popen('cd %s;git tag' % dd_folder, stdout=subprocess.PIPE, shell=True).communicate()[0])
    tags = list(filter(lambda x: str(x).startswith('3.') and int(x.split('.')[1]) >= 10, result.split()))
    # add development branch at the beginning of list of tags
    tags.insert(0, 'develop/3')
    imas_versions = OrderedDict()
    for item in tags:
        imas_versions[item] = item.replace('.', '_').replace('/', '_')
    if imas_version is None:
        return imas_versions

    # generate IDSDef.xml files for given imas_version
    _imas_version = imas_versions[imas_version]
    executable = """
export CLASSPATH={imas_json_dir}/../{saxon_version}/saxon9he.jar;
cd {dd_folder}
git checkout {tag}
git pull
export JAVA_HOME=$(dirname $(dirname `which java`))
make clean
make
rm -rf {imas_json_dir}/{_imas_version}/
mkdir {imas_json_dir}/{_imas_version}/
cp IDSDef.xml {imas_json_dir}/{_imas_version}/
""".format(tag=imas_version, _imas_version=_imas_version, imas_json_dir=imas_json_dir, dd_folder=dd_folder, saxon_version=saxon_version)
    print(executable)
    subprocess.Popen(executable, shell=True).communicate()


def create_json_structure(imas_version=omas_rcparams['default_imas_version']):
    import xmltodict
    file = imas_json_dir + os.sep + imas_versions.get(imas_version, imas_version) + os.sep + 'IDSDef.xml'
    tmp = xmltodict.parse(open(file).read())

    for file in glob.glob(imas_json_dir + os.sep + imas_versions.get(imas_version, imas_version) + os.sep + '*.json'):
        print('Remove ' + file)
        os.remove(file)

    def process_path(inv):
        inv = inv.replace(r'/', '.')
        inv = inv.split(' OR ')[0]
        inv = remove_parentheses(inv, '[:]')
        inv = re.sub('\[:\]$', '', inv)
        return inv

    def traverse(me, hout, path, fout, parent):
        me = copy.copy(me)
        hout_propagate = hout
        path_propagate = copy.deepcopy(path)
        parent = copy.deepcopy(parent)

        if '@structure_reference' in me and me['@structure_reference'] == 'self':
            return hout, fout

        if '@name' in me:
            name = me['@name']
            path_propagate.append(name)
            hout_propagate = {}
            hout[name] = hout_propagate
            # paths
            if '@path_doc' in me:
                fname = path_propagate[0] + '/' + me['@path_doc']
                me['@full_path'] = path_propagate[0] + '/' + me['@path_doc']
                del me['@path_doc']
            else:
                fname = path_propagate[0]
                me['@full_path'] = path_propagate[0]
            if '@path' in me:
                del me['@path']
            # coordinates
            for coord in [c for c in me if c.startswith('@coordinate')]:
                if '...' not in me[coord]:
                    me[coord] = path_propagate[0] + '/' + me[coord]
            # identifiers documentation
            if '@doc_identifier' in me:
                doc_id = xmltodict.parse(open(imas_json_dir + '/../data-dictionary/' + me['@doc_identifier']).read())
                hlp = doc_id['constants']['int']
                doc = []
                for row in hlp:
                    doc.append('%s) %s : %s' % (row['#text'], row['@name'], row['@description']))
                me['@documentation'] = me['@documentation'].strip() + '\n' + '\n'.join(doc)

            fname = process_path(fname)
            fout[fname] = {}

        if isinstance(me, list):
            keys = range(len(me))
        else:
            keys = list(me.keys())

        if '@units' in me:
            if fname == 'equilibrium.time_slice[:].constraints.q':  # bug fix for v3.18.0
                me['@units'] = '-'
            if fname in ['equilibrium.time_slice[:].profiles_1d.geometric_axis.r', 'equilibrium.time_slice[:].profiles_1d.geometric_axis.z']:
                me['@coordinate'] = ['equilibrium.time_slice[:].profiles_1d.psi']
            if me['@units'] in ['as_parent', 'as parent', 'as_parent_level_2']:
                me['@units'] = parent['units']
            parent['units'] = me['@units']

        # children inherit lifecycle status from parent
        if '@lifecycle_status' in me:
            parent['lifecycle_status'] = me['@lifecycle_status']
        elif parent['lifecycle_status'] and not isinstance(me, list):
            me['@lifecycle_status'] = parent['lifecycle_status']
            keys.append('@lifecycle_status')

        is_leaf = True
        for kid in keys:
            if isinstance(me[kid], (dict, list)):
                is_leaf = False
                traverse(me[kid], hout_propagate, path_propagate, fout, parent)
            elif kid not in ['@name', '@xmlns:fn']:
                hout_propagate[kid] = me[kid]
                fout[fname][kid] = me[kid]

                # if is_leaf:
                # print(path_propagate)

        return hout, fout

    parent = {}
    parent['units'] = '?'
    parent['lifecycle_status'] = ''
    hout, fout = traverse(me=tmp, hout={}, path=[], fout={}, parent=parent)

    # format conversions
    for item in sorted(fout):
        coords = []
        for key in sorted(list(fout[item].keys())):
            if key.endswith('AosParent_relative') or key.endswith('_same_as'):
                del fout[item][key]
            elif key != '@coordinates' and key.startswith('@coordinate'):
                coords.append(fout[item][key])
                del fout[item][key]
            elif key.startswith('@path'):
                fout[item][key] = process_path(fout[item][key])
        if len(coords):
            # # this is a check for duplicated coordinates, which it is not an error per se
            # if len(numpy.unique(list(filter(lambda x: not x.startswith('1...'), coords)))) != len(list(filter(lambda x: not x.startswith('1...'), coords))):
            #     printe('%s -X-> %s' % (item, coords))
            coords = list(map(process_path, coords))
            fout[item]['@coordinates'] = coords

    # deal with cross-referenced coordinates
    for item in sorted(fout):
        if '@coordinates' in fout[item]:
            fout[item]['@coordinates'] = list(map(lambda x: re.sub('^.*\.IDS:', '', x), fout[item]['@coordinates']))

    # check dimensions
    for item in sorted(fout):
        coord = []
        if '@coordinates' in fout[item]:
            for coord in fout[item]['@coordinates']:
                if '...' in coord:
                    continue
                if coord not in fout:
                    printe('%s --> %s' % (item, coord))

    # cleanup entries
    for item in sorted(fout):
        keys = fout[item].keys()
        values = fout[item].values()
        keys = list(map(lambda x: re.sub('^@', '', x), keys))
        fout[item] = dict(zip(keys, values))

    # break into pieces
    hout = {}
    for item in sorted(fout):
        #    if re.findall('(_error_upper|_error_lower|_error_index)$', item):
        #        continue
        ds = item.split('.')[0]
        hout.setdefault(ds, {})[item] = fout[item]

    # additional data structures
    hout.update(add_datastructures)

    # prepare directory structure
    if not os.path.exists(imas_json_dir + os.sep + imas_versions.get(imas_version, imas_version)):
        os.makedirs(imas_json_dir + os.sep + imas_versions.get(imas_version, imas_version))
    for item in glob.glob(imas_json_dir + os.sep + imas_versions.get(imas_version, imas_version) + os.sep + '*.json'):
        os.remove(item)

    # deploy imas structures as json
    for structure in sorted(hout):
        if structure == 'time':
            continue
        print(imas_json_dir + os.sep + imas_versions.get(imas_version, imas_version) + os.sep + structure + '.json')
        dump_string = json.dumps(hout[structure], default=json_dumper, indent=1, separators=(',', ': '), sort_keys=True)
        open(imas_json_dir + os.sep + imas_versions.get(imas_version, imas_version) + os.sep + structure + '.json', 'w').write(dump_string)

    # generate coordinates cache file
    coords = extract_coordinates(imas_version=imas_version)
    dump_string = json.dumps(coords, default=json_dumper, indent=1, separators=(',', ': '), sort_keys=True)
    open(imas_json_dir + os.sep + imas_versions.get(imas_version, imas_version) + os.sep + '_coordinates.json', 'w').write(dump_string)


def create_html_documentation(imas_version=omas_rcparams['default_imas_version']):
    filename = os.path.abspath(os.sep.join([imas_json_dir, imas_versions.get(imas_version, imas_version), 'omas_doc.html']))

    table_header = "<table border=1, width='100%'>"
    sub_table_header = '<tr>' \
                       '<th style="width:25%">Path</th>' \
                       '<th style="width:25%">Dimensions</th>' \
                       '<th>Type</th>' \
                       '<th>Units</th>' \
                       '<th>Description</th>' \
                       '</tr>'

    column_style = 'style="word-wrap:break-word;word-break:break-all"'
    lines = []
    for structure_file in list_structures(imas_version=imas_version):
        print('Adding to html documentation: ' + structure_file)
        # load structure information
        structure = load_structure(structure_file, imas_version=imas_version)[0]
        # idenfity coordinates in the structure
        coords = []
        for item in structure:
            if 'coordinates' in structure[item]:
                coords.extend(structure[item]['coordinates'])
        coords = list(filter(lambda x: '...' not in x, set(coords)))
        # generate output
        lines.append('<!-- %s -->' % structure_file)
        lines.append(table_header)
        lines.append(sub_table_header)
        for item in sorted(structure):
            if not any([item.endswith(k) for k in ['_error_index', '_error_lower', '_error_upper']]):
                # uncertain quantities
                is_uncertain = ''
                if item + '_error_upper' in structure:
                    is_uncertain = ' (uncertain)'
                # lifecycle status
                status = ''
                if 'lifecycle_status' in structure[item] and structure[item]['lifecycle_status'] not in ['active']:
                    color_mapper = {'alpha': 'blue', 'obsolescent': 'red'}
                    status = '</p><p><font color="%s">(%s)</font>' % (color_mapper.get(structure[item]['lifecycle_status'], 'orange'), structure[item]['lifecycle_status'])
                # highlight entries that are a coordinate
                item_with_coordinate_highlight = item
                if item in coords:
                    item_with_coordinate_highlight = '<strong>%s</strong>' % item
                try:
                    lines.append(
                        '<tr>'
                        '<td {column_style}><p>{item_with_coordinate_highlight}{status}</p></td>' \
                        '<td {column_style}><p>{coordinates}</p></td>' \
                        '<td><p>{data_type}</p></td>' \
                        '<td><p>{units}</p></td>' \
                        '<td><p>{description}</p></td>'
                        '</tr>'.format(
                            item_with_coordinate_highlight=item_with_coordinate_highlight,
                            coordinates=re.sub('\[\]', '', re.sub('[\'\"]', '', re.sub(',', ',<br>', str(
                                list(map(str, structure[item].get('coordinates', ''))))))),
                            data_type=structure[item].get('data_type', '') + is_uncertain,
                            units=structure[item].get('units', ''),
                            description=re.sub('\n', '<br>', structure[item].get('documentation', '')),
                            status=status,
                            column_style=column_style
                        ))
                except Exception:
                    printe(item)
                    raise
        lines.append('</table>')
        lines.append('</table><p></p>')

    with open(filename, 'w') as f:
        if sys.version_info < (3, 0):
            f.write('\n'.join(lines).encode('utf-8'))
        else:
            f.write('\n'.join(lines))


def extract_coordinates(imas_version=omas_rcparams['default_imas_version']):
    '''
    return list of strings with coordinates across all structures

    :param imas_version: imas version

    :return: list with coordinate
    '''
    from omas.omas_utils import list_structures
    from omas.omas_utils import load_structure

    omas_coordinates = set()
    for structure in list_structures(imas_version=imas_version):
        tmp = load_structure(structure, imas_version)[0]
        coords = []
        for item in tmp:
            if 'coordinates' in tmp[item]:
                coords.extend(map(i2o, tmp[item]['coordinates']))
        coords = list(filter(lambda x: '...' not in x, set(coords)))
        omas_coordinates.update(coords)

    return sorted(list(omas_coordinates))


def symlink_imas_structure_versions(test=True, verbose=True):
    '''
    Generate symbolic links in imas_structures so that no files are added when there are no changes between IDSs

    :param test: whether to actually apply symlink commands

    :param verbose: wheter to print to screen structures strides

    :returns: dictionary with structure stides per IDS
    '''

    import subprocess
    from pprint import pprint
    from omas.omas_setup import IMAS_versions
    from omas.omas_utils import dict_structures, imas_json_dir

    imas_versions = IMAS_versions('tagged')

    # check if two files are identical
    def same_ds(a, b):
        with open(a) as a:
            with open(b) as b:
                return a.read() == b.read()

    # see when DS files have changed
    structures_strides = {}
    previous_tag_structures = {}
    for version in list(imas_versions.keys()):
        this_tag_structures = dict_structures(version)
        for ds in this_tag_structures:
            if ds not in structures_strides:
                structures_strides[ds] = [[version]]
            elif ds in previous_tag_structures:
                if same_ds(previous_tag_structures[ds], this_tag_structures[ds]):
                    structures_strides[ds][-1].append(version)
                else:
                    structures_strides[ds].append([version])
        previous_tag_structures = this_tag_structures
    if verbose:
        pprint(structures_strides)

    if not test:
        for ds in structures_strides:
            for stride in structures_strides[ds]:
                if len(stride) > 1:
                    for version in stride[:-1]:
                        dir = imas_json_dir + '/'
                        this = dict_structures(stride[-1])[ds][len(dir):]
                        prev = dict_structures(version)[ds]
                        command = 'cd %s; ln -s -f ../%s %s' % (os.path.dirname(prev), this, os.path.basename(prev))
                        print(command)
                        subprocess.Popen(command, shell=True).communicate()
    return structures_strides